
import pytest
from unittest.mock import MagicMock, patch
from veeksha.config.generator.channel import TextChannelGeneratorConfig
from veeksha.config.generator.length import FixedLengthGeneratorConfig
from veeksha.generator.channel.text import TextChannelGenerator
from veeksha.core.seeding import SeedManager

@pytest.fixture
def mock_tokenizer():
    tokenizer = MagicMock()
    tokenizer.encode.side_effect = lambda x: [1] * len(x.split())  # simple mockery
    tokenizer.decode.side_effect = lambda x: " ".join(["word"] * len(x))
    tokenizer.count_tokens.side_effect = lambda x: len(x.split()) if isinstance(x, str) else len(x)
    return tokenizer

@pytest.fixture
def mock_seed_manager():
    return SeedManager(seed=42)

@pytest.fixture
def mock_load_corpus():
    with patch("veeksha.generator.channel.text.load_corpus") as mock:
        mock.return_value = ["line one", "line two", "line three"]
        yield mock

@pytest.fixture
def text_config():
    return TextChannelGeneratorConfig(
        body_length_generator=FixedLengthGeneratorConfig(value=10),
        output_length_generator=FixedLengthGeneratorConfig(value=5),
        shared_prefix_ratio=0.5,
        shared_prefix_probability=1.0,  # Always use shared prefix if is_root
    )

def test_text_generator_initialization(mock_load_corpus, mock_tokenizer, mock_seed_manager, text_config):
    generator = TextChannelGenerator(text_config, mock_seed_manager, mock_tokenizer)
    assert generator.config == text_config
    assert len(generator._corpus_lines) == 3

def test_generate_content_basic(mock_load_corpus, mock_tokenizer, mock_seed_manager, text_config):
    generator = TextChannelGenerator(text_config, mock_seed_manager, mock_tokenizer)
    content = generator.generate_content(is_root=False)
    
    # Check if target lengths match config
    assert content.target_prompt_tokens == 10
    assert content.target_output_tokens == 5
    assert isinstance(content.input_text, str)

def test_generate_content_shared_prefix(mock_load_corpus, mock_tokenizer, mock_seed_manager, text_config):
    # Config has 0.5 ratio and 1.0 probability
    generator = TextChannelGenerator(text_config, mock_seed_manager, mock_tokenizer)
    
    # First root request
    content1 = generator.generate_content(is_root=True)
    
    # Second root request
    content2 = generator.generate_content(is_root=True)
    
    # With fixed length 10 and ratio 0.5, prefix length is 5.
    # The generator should use the same prefix for both.
    # Since we rely on random corpus generation, exact text match depends on seeding.
    # However, _generate_shared_prefix caches the prefix.
    
    # We can check internal state or verifying that _generate_shared_prefix was called and used cached tokens
    # But better to check if the logic holds.
    
    assert len(generator._shared_prefix_tokens) >= 5

def test_generate_content_min_tokens_instruction(mock_load_corpus, mock_tokenizer, mock_seed_manager, text_config):
    # Enable instruction
    generator = TextChannelGenerator(
        text_config, 
        mock_seed_manager, 
        mock_tokenizer, 
        append_min_tokens_instruction=True
    )
    
    # Mock tokenizer to return short length for suffix so it fits
    mock_tokenizer.encode.side_effect = lambda x: [1] * (1 if "Generate" in x else len(x.split()))
    
    content = generator.generate_content(is_root=False)
    assert "Generate at least" in content.input_text or content.target_prompt_tokens == 10



"""Generate Sphinx documentation for veeksha configuration.

This module generates RST files for the Sphinx documentation from the config schema.
Each dataclass gets its own RST file with a readable pseudo-code format.
"""

from __future__ import annotations

import re
from dataclasses import MISSING, fields, is_dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, get_args, get_origin

from veeksha.config.core.base_poly_config import BasePolyConfig
from veeksha.config.utils import get_all_subclasses, is_optional


def _to_human_readable(class_name: str) -> str:
    """Convert a class name to human-readable format.

    E.g., 'AudioChannelGeneratorConfig' → 'Audio Channel Generator'
         'OpenAIChatCompletionsClientConfig' → 'OpenAI Chat Completions Client'
         'RAGTraceFlavorConfig' → 'RAG Trace Flavor'
    """
    # Remove 'Config' suffix
    name = class_name
    if name.endswith("Config"):
        name = name[:-6]

    # Use regex to insert spaces at camelCase boundaries
    # This splits before a capital that's followed by lowercase OR
    # before the last capital in a sequence of capitals
    result = re.sub(r"(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])", " ", name)

    # Fix known brand names that may have been incorrectly split
    fixes = {
        "Open AI": "OpenAI",
        "LM Eval": "LMEval",
        "Lm eval": "LMEval",
        # These don't get split by the regex since they have no internal capitals
        "Vllm": "vLLM",
        "Sglang": "SGLang",
    }
    for wrong, correct in fixes.items():
        result = result.replace(wrong, correct)

    return result


def generate_sphinx_docs(output_dir: str = "docs/config_reference") -> None:
    """Generate Sphinx RST documentation for all config types."""
    from veeksha.config.benchmark import BenchmarkConfig
    from veeksha.config.capacity_search import CapacitySearchConfig

    output_path = Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)

    # Create api_reference subdirectory
    api_path = output_path / "api_reference"
    api_path.mkdir(parents=True, exist_ok=True)

    # Collect all dataclasses from root configs
    all_classes: Dict[str, Type] = {}
    _collect_dataclasses(BenchmarkConfig, all_classes)
    _collect_dataclasses(CapacitySearchConfig, all_classes)

    # Generate RST for each class
    for class_name, cls in sorted(all_classes.items()):
        rst_content = _generate_class_rst(cls, all_classes)
        (api_path / f"{class_name}.rst").write_text(rst_content)

    # Generate API reference index
    api_index = _generate_api_index(all_classes)
    (api_path / "index.rst").write_text(api_index)

    # Generate main config reference index
    main_index = _generate_main_index()
    (output_path / "index.rst").write_text(main_index)

    print(f"Generated API reference for {len(all_classes)} classes in {api_path}")


def _collect_dataclasses(cls: Type, collected: Dict[str, Type]) -> None:
    """Recursively collect all dataclass types from a root class."""
    if not is_dataclass(cls):
        return

    class_name = cls.__name__
    if class_name in collected:
        return

    collected[class_name] = cls

    # Collect from fields
    for f in fields(cls):
        field_type = f.type
        _collect_from_type(field_type, collected)


def _collect_from_type(field_type: Type, collected: Dict[str, Type]) -> None:
    """Extract dataclass types from a type annotation."""
    # Handle Optional[T]
    if is_optional(field_type):
        args = get_args(field_type)
        for arg in args:
            if arg is not type(None):
                _collect_from_type(arg, collected)
        return

    # Handle List[T]
    origin = get_origin(field_type)
    if origin is list:
        args = get_args(field_type)
        if args:
            _collect_from_type(args[0], collected)
        return

    # Handle polymorphic base classes
    if isinstance(field_type, type) and issubclass(field_type, BasePolyConfig):
        _collect_dataclasses(field_type, collected)
        # Also collect all subclasses
        for subclass in get_all_subclasses(field_type):
            _collect_dataclasses(subclass, collected)
        return

    # Handle regular dataclasses
    if is_dataclass(field_type):
        _collect_dataclasses(field_type, collected)


def _generate_class_rst(cls: Type, all_classes: Dict[str, Type]) -> str:
    """Generate RST documentation for a single dataclass."""
    lines: List[str] = []
    class_name = cls.__name__

    # Title
    lines.append(class_name)
    lines.append("=" * len(class_name))
    lines.append("")

    # Class signature
    parent_classes = [
        base.__name__
        for base in cls.__bases__
        if base.__name__ != "object" and is_dataclass(base)
    ]

    if parent_classes:
        parent_links = ", ".join(
            f":doc:`{p}`" if p in all_classes else f"``{p}``" for p in parent_classes
        )
        lines.append(f"**class** ``{class_name}`` **(** {parent_links} **)**")
    else:
        lines.append(f"**class** ``{class_name}``")
    lines.append("")

    # Docstring - skip auto-generated constructor signatures
    if cls.__doc__:
        docstring = cls.__doc__.strip()
        # Skip if it looks like an auto-generated dataclass signature
        if not (docstring.startswith(class_name + "(") and "=" in docstring):
            # Extract only first paragraph (stop at Attributes:, Args:, etc.)
            first_para = _extract_first_paragraph(docstring)
            if first_para:
                lines.append(f"    {first_para}")
                lines.append("")

    # Polymorphic type info (if this is a poly subclass)
    if (
        isinstance(cls, type)
        and issubclass(cls, BasePolyConfig)
        and cls is not BasePolyConfig
    ):
        has_header = False

        # 1. Try to display the specific type of this class
        try:
            type_value = cls.get_type()
            if hasattr(type_value, "name"):
                type_value = type_value.name.lower()
            lines.append("**Polymorphic Type:**")
            lines.append("")
            lines.append(f"    ``type: {type_value}``")
            has_header = True
        except NotImplementedError:
            pass

        # 2. Display all available types for this family
        # Find the root polymorphic base (closest to BasePolyConfig)
        mro = cls.mro()
        # Filter for custom classes that inherit from BasePolyConfig
        poly_parents = [
            c
            for c in mro
            if issubclass(c, BasePolyConfig)
            and c is not BasePolyConfig
            and c is not object
        ]

        variant_lines = []
        if poly_parents:
            poly_root = poly_parents[-1]
            variant_lines = _get_variant_lines(poly_root, all_classes)

            if variant_lines:
                if not has_header:
                    lines.append("**Polymorphic Type:**")
                    lines.append("")
                else:
                    lines.append("")

                lines.append(f"    All ``{poly_root.__name__}`` types:")
                lines.append("")
                for v_line in variant_lines:
                    lines.append(f"    {v_line}")

        # Add trailing newline if we added any polymorphic info
        if has_header or (poly_parents and variant_lines):
            lines.append("")

    # Fields section
    class_fields = list(fields(cls))
    if class_fields:
        lines.append("**Fields:**")
        lines.append("")

        for f in class_fields:
            if f.name == "type":  # Skip polymorphic type discriminator
                continue
            _generate_field_rst(f, lines, all_classes)

    return "\n".join(lines)


def _generate_field_rst(f: Any, lines: List[str], all_classes: Dict[str, Type]) -> None:
    """Generate RST for a single field."""
    field_name = f.name
    field_type = f.type
    help_text = f.metadata.get("help", "") if f.metadata else ""

    # Get type string and check for polymorphic
    type_str, is_poly, poly_base = _get_type_info(field_type, all_classes)

    # Get default value
    default_str = _get_default_str(f)

    # Field line
    lines.append(f"``{field_name}`` : {type_str} = {default_str}")

    # Help text
    if help_text:
        # Clean up help text (remove "Available: ..." since we show it below)
        clean_help = help_text.split(". Available:")[0].strip()
        if clean_help:
            lines.append(f"    {clean_help}")

    # Available types for polymorphic fields
    if is_poly and poly_base:
        variant_lines = _get_variant_lines(poly_base, all_classes)
        if variant_lines:
            lines.append("")
            lines.append("    Available types:")
            lines.append("")
            for v_line in variant_lines:
                lines.append(f"    {v_line}")

    lines.append("")


def _get_type_info(
    field_type: Type, all_classes: Dict[str, Type]
) -> tuple[str, bool, Optional[Type]]:
    """Get type string, whether it's polymorphic, and the polymorphic base if so."""
    actual_type = field_type
    is_optional_type = False
    is_list_type = False

    # Handle Optional[T]
    if is_optional(field_type):
        is_optional_type = True
        args = get_args(field_type)
        actual_type = next((t for t in args if t is not type(None)), field_type)

    # Handle List[T]
    origin = get_origin(actual_type)
    if origin is list:
        is_list_type = True
        args = get_args(actual_type)
        if args:
            actual_type = args[0]

    # Check if polymorphic
    is_poly = False
    poly_base = None
    if isinstance(actual_type, type) and issubclass(actual_type, BasePolyConfig):
        is_poly = True
        poly_base = actual_type

    # Build type string
    if isinstance(actual_type, type):
        type_name = actual_type.__name__
        if type_name in all_classes:
            type_str = f":doc:`{type_name}`"
        else:
            type_str = f"*{type_name}*"
    else:
        type_str = f"*{actual_type}*"

    if is_list_type:
        type_str = f"*list* [ {type_str} ]"
    if is_optional_type:
        type_str = f"*Optional* [ {type_str} ]"

    return type_str, is_poly, poly_base


def _get_default_str(f: Any) -> str:
    """Get default value string for a field."""
    if f.default is not MISSING:
        return _format_default(f.default)
    if f.default_factory is not MISSING:
        try:
            value = f.default_factory()
            return _format_default(value)
        except Exception:
            return "``<factory>``"
    return "*required*"


def _format_default(value: Any) -> str:
    """Format a default value for display."""
    if value is MISSING:
        return "*required*"
    if isinstance(value, type):
        return f":doc:`{value.__name__}`"
    if hasattr(value, "__dataclass_fields__"):
        return f":doc:`{value.__class__.__name__}`"
    if hasattr(value, "name"):  # Enum
        return f"``{value.name.lower()}``"
    if isinstance(value, str):
        return f'``"{value}"``'
    if isinstance(value, bool):
        return f"``{value}``"
    if isinstance(value, (int, float)):
        return f"``{value}``"
    if isinstance(value, list):
        if not value:
            return "``[]``"
        # For lists of dataclasses, show simplified form
        if value and hasattr(value[0], "__dataclass_fields__"):
            return f"``[<{value[0].__class__.__name__}>]``"
        # Keep short lists, simplify long ones
        str_repr = str(value)
        if len(str_repr) > 50:
            return f"``[...]``"
        return f"``{value}``"
    # For any complex object, simplify
    str_repr = str(value)
    if len(str_repr) > 50:
        return f"``<{type(value).__name__}>``"
    return f"``{value}``"


def _extract_first_paragraph(docstring: str) -> str:
    """Extract only the first paragraph from a docstring.

    Stops at section headers like Attributes:, Args:, Returns:, etc.
    """
    lines = docstring.split("\n")
    result_lines = []

    for line in lines:
        stripped = line.strip()
        # Stop at section headers (commonly used in docstrings)
        if stripped and stripped.endswith(":") and stripped[:-1].isalpha():
            break
        # Stop at blank line after we have content
        if not stripped and result_lines:
            break
        if stripped:
            result_lines.append(stripped)

    return " ".join(result_lines)


def _get_variant_lines(
    poly_base: Type[BasePolyConfig], all_classes: Dict[str, Type]
) -> List[str]:
    """Get formatted list of polymorphic variants."""
    subclasses = get_all_subclasses(poly_base)
    lines = []
    for subclass in subclasses:
        class_name = subclass.__name__
        try:
            type_value = subclass.get_type()
            if hasattr(type_value, "name"):
                type_value = type_value.name.lower()
        except NotImplementedError:
            type_value = class_name.lower()

        if class_name in all_classes:
            link = f":doc:`{class_name}`"
        else:
            link = f"``{class_name}``"

        lines.append(f"- ``{type_value}``: {link}")

    return lines


def _generate_api_index(all_classes: Dict[str, Type]) -> str:
    """Generate the API reference index page with grouped classes."""
    # Import base classes for grouping
    from veeksha.config.client import BaseClientConfig
    from veeksha.config.evaluator import (
        BaseChannelPerformanceConfig,
        BaseEvaluatorConfig,
    )
    from veeksha.config.generator.channel import BaseChannelGeneratorConfig
    from veeksha.config.generator.interval import BaseIntervalGeneratorConfig
    from veeksha.config.generator.length import BaseLengthGeneratorConfig
    from veeksha.config.generator.session import (
        BaseSessionGeneratorConfig,
        BaseTraceFlavorConfig,
    )
    from veeksha.config.generator.session_graph import BaseSessionGraphGeneratorConfig
    from veeksha.config.server import BaseServerConfig
    from veeksha.config.slo import BaseSloConfig
    from veeksha.config.traffic import BaseTrafficConfig

    # Group definitions: (Section Title, Base Class or None for root, classes list)
    groups: Dict[str, List[str]] = {
        "Root configurations": [],
        "Session generators": [],
        "Traffic schedulers": [],
        "Clients": [],
        "Servers": [],
        "Evaluators": [],
        "Channel performance": [],
        "Length generators": [],
        "Interval generators": [],
        "Channel generators": [],
        "Session graph generators": [],
        "Trace flavors": [],
        "SLO configs": [],
        "Other": [],
    }

    # Mapping of base classes to group names
    base_to_group: List[tuple[Type, str]] = [
        (BaseSessionGeneratorConfig, "Session generators"),
        (BaseTrafficConfig, "Traffic schedulers"),
        (BaseClientConfig, "Clients"),
        (BaseServerConfig, "Servers"),
        (BaseEvaluatorConfig, "Evaluators"),
        (BaseChannelPerformanceConfig, "Channel performance"),
        (BaseLengthGeneratorConfig, "Length generators"),
        (BaseIntervalGeneratorConfig, "Interval generators"),
        (BaseChannelGeneratorConfig, "Channel generators"),
        (BaseSessionGraphGeneratorConfig, "Session graph generators"),
        (BaseTraceFlavorConfig, "Trace flavors"),
        (BaseSloConfig, "SLO configs"),
    ]

    # Root configs
    root_configs = {"BenchmarkConfig", "CapacitySearchConfig"}

    # Classify each class
    for class_name, cls in all_classes.items():
        if class_name in root_configs:
            groups["Root configurations"].append(class_name)
            continue

        # Check against base classes
        found = False
        for base_cls, group_name in base_to_group:
            if isinstance(cls, type) and issubclass(cls, base_cls):
                groups[group_name].append(class_name)
                found = True
                break

        if not found:
            groups["Other"].append(class_name)

    # Build the RST content
    lines = [
        "API Reference",
        "=============",
        "",
        "This section contains the API reference for all configuration classes, organized by category.",
        "",
    ]

    # Generate sections
    for group_name, class_names in groups.items():
        if not class_names:
            continue

        # Section header
        lines.append(group_name)
        lines.append("-" * len(group_name))
        lines.append("")

        # Root configs get special treatment with descriptions
        if group_name == "Root configurations":
            lines.append(
                "- :doc:`BenchmarkConfig` - Configuration for ``veeksha.benchmark`` runs"
            )
            lines.append(
                "- :doc:`CapacitySearchConfig` - Configuration for ``veeksha.capacity_search`` runs"
            )
            lines.append("")
            continue

        # Toctree with human-readable display names
        lines.append(".. toctree::")
        lines.append("   :maxdepth: 1")
        lines.append("")

        for class_name in sorted(class_names):
            display_name = _to_human_readable(class_name)
            lines.append(f"   {display_name} <{class_name}>")

        lines.append("")

    return "\n".join(lines)


def _generate_main_index() -> str:
    """Generate the main config reference index."""
    return """Configuration Reference
=======================

This section provides a comprehensive reference for all configuration options in Veeksha.
Configuration can be provided via YAML files or CLI arguments.

.. tip::
   
   Use the interactive config explorer for an easier experience::
   
       python -m veeksha.cli.config explore

   Or generate a YAML schema template::
   
       python -m veeksha.cli.config show --format yaml


Quick links
-----------

- :doc:`api_reference/BenchmarkConfig` - Configuration for standard benchmark runs
- :doc:`api_reference/CapacitySearchConfig` - Configuration for capacity search experiments
- :doc:`api_reference/index` - Full API reference for all config classes


Understanding the config system
-------------------------------

Veeksha uses a **polymorphic configuration system**. Many options have a ``type`` field
that determines which variant is used, each with its own set of options.

For example, the ``traffic_scheduler`` can be either ``rate`` or ``concurrent``::

    # Rate-based traffic
    traffic_scheduler:
      type: rate
      interval_generator:
        type: poisson
        rate: 10.0  # 10 requests per second

    # Concurrency-based traffic
    traffic_scheduler:
      type: concurrent
      target_concurrent_sessions: 8
      rampup_seconds: 10


IDE autocompletion
------------------

See :doc:`/basic_usage/configuration` for instructions on how to set up YAML autocompletion and validation in your IDE.


.. toctree::
   :maxdepth: 2
   :hidden:

   api_reference/index
"""


if __name__ == "__main__":
    generate_sphinx_docs()



from abc import ABC, abstractmethod

from veeksha.config.generator.length import (
    BaseLengthGeneratorConfig,
)


class BaseLengthGenerator(ABC):
    def __init__(self, config: BaseLengthGeneratorConfig, rng=None):
        self.config = config
        self.rng = rng

    @abstractmethod
    def get_next_value(self) -> int:
        pass



"""
Context manager for automatic server lifecycle management.
"""

import os
from contextlib import contextmanager
from typing import Any, Dict, Generator

from veeksha.config.server import BaseServerConfig
from veeksha.orchestration.registry import ServerManagerRegistry
from veeksha.orchestration.server_manager import BaseServerManager


def create_server_manager(
    config: BaseServerConfig,
    output_dir: str,
) -> BaseServerManager:
    """Create appropriate server manager based on config type."""
    return ServerManagerRegistry.get(
        config.get_type(),
        config=config,
        output_dir=output_dir,
    )


@contextmanager
def managed_server(
    config: BaseServerConfig,
    output_dir: str,
) -> Generator[Dict[str, Any], None, None]:
    """Context manager for automatic server lifecycle management.

    Handles:
    1. Launch server
    2. Wait for ready
    3. Yield server info
    4. Shutdown

    Args:
        config: Server configuration
        output_dir: Directory for server logs.

    Yields:
        Dictionary with server info:
            - api_base: API base URL
            - api_key: API key
            - server_manager: Server manager instance
    """
    server_manager = create_server_manager(config, output_dir=output_dir)

    try:
        launch_result = server_manager.launch()
        if isinstance(launch_result, tuple):
            success, error = launch_result
        else:
            success, error = bool(launch_result), None
        if not success:
            raise RuntimeError(f"Failed to launch server: {error}")

        if not server_manager.wait_for_ready():
            raise RuntimeError("Server failed to become ready")

        api_base = config.get_api_base_url()

        # Set environment variables for clients
        os.environ["OPENAI_API_KEY"] = config.api_key or "EMPTY"
        os.environ["OPENAI_API_BASE"] = api_base

        yield {
            "api_base": api_base,
            "api_key": config.api_key,
            "server_manager": server_manager,
        }

    finally:
        server_manager.shutdown()



from __future__ import annotations

import math
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple

from lm_eval.api.instance import (
    Instance,
)
from lm_eval.api.task import Task
from lm_eval.evaluator_utils import (
    get_task_list,
)
from lm_eval.tasks import (
    TaskManager,
    get_task_dict,
)

from veeksha.config.generator.session import LmevalSessionGeneratorConfig
from veeksha.core.request import Request
from veeksha.core.request_content import (
    TextChannelRequestContent,
)
from veeksha.core.seeding import SeedManager
from veeksha.core.session import Session
from veeksha.core.session_graph import SessionGraph, SessionNode, add_node
from veeksha.core.tokenizer import TokenizerProvider
from veeksha.generator.session.base import BaseSessionGenerator
from veeksha.logger import init_logger
from veeksha.types import ChannelModality, LMEvalOutputType

logger = init_logger(__name__)


class LMEvalSessionGenerator(BaseSessionGenerator):
    """Session generator that emits one session per lm-eval document.

    The generator owns lm-eval Task/Instance objects. The accuracy evaluator binds
    responses back to these instances using request_id.

    Note:
        A single lm-eval document can require multiple model calls (e.g., multiple
        loglikelihood instances for multiple-choice). In those cases, the generator
        emits a single Session containing multiple Requests.
    """

    def __init__(
        self,
        config: LmevalSessionGeneratorConfig,
        seed_manager: SeedManager,
        tokenizer_provider: TokenizerProvider,
        max_sessions: int = -1,
    ):
        self.config = config
        self.seed_manager = seed_manager
        self.tokenizer_provider = tokenizer_provider
        self.text_tokenizer = self.tokenizer_provider.for_modality(ChannelModality.TEXT)
        self.max_sessions = max_sessions

        self._fewshot_rng = self.seed_manager.random("fewshot")

        self.task_manager = TaskManager()
        self.task_dict = get_task_dict(self.config.tasks, self.task_manager)  # type: ignore
        if not self.task_dict:
            raise ValueError("LMEvalSessionGenerator could not resolve any tasks.")

        self.task_dict = self._adjust_config(self.task_dict)

        self.eval_tasks = get_task_list(self.task_dict)
        num_tasks = max(1, len(self.eval_tasks))
        per_task_doc_limit: Optional[int]
        if self.max_sessions < 0:
            per_task_doc_limit = None
        else:
            per_task_doc_limit = max(0, int(math.ceil(self.max_sessions / num_tasks)))

        # Build and collect per-task doc ordering.
        task_to_doc_ids: Dict[str, List[int]] = {}
        task_to_instances_by_doc: Dict[str, Dict[int, List[Instance]]] = {}

        for task_output in self.eval_tasks:
            task: Task = task_output.task  # type: ignore
            task.build_all_requests(limit=per_task_doc_limit)

            instances_by_doc_id: Dict[int, List[Instance]] = defaultdict(list)
            for instance in task.instances:
                doc_id = instance.doc_id
                if doc_id is None:
                    continue
                instances_by_doc_id[doc_id].append(instance)
            for instances in instances_by_doc_id.values():
                instances.sort(key=lambda x: x.idx)

            tname = str(task_output.task_name)
            doc_ids: List[int] = [
                doc_id
                for doc_id, _ in task.doc_iterator(limit=per_task_doc_limit)
                if doc_id is not None
            ]
            task_to_doc_ids[tname] = doc_ids
            task_to_instances_by_doc[tname] = instances_by_doc_id

        # Interleave tasks' doc streams (round-robin) and then truncate to max_sessions.
        all_doc_keys: List[Tuple[str, int]] = []
        max_len = max((len(v) for v in task_to_doc_ids.values()), default=0)
        for i in range(max_len):
            for task_output in self.eval_tasks:
                tname = str(task_output.task_name)
                doc_ids = task_to_doc_ids.get(tname, [])
                if i < len(doc_ids):
                    all_doc_keys.append((tname, doc_ids[i]))

        if self.max_sessions >= 0:
            all_doc_keys = all_doc_keys[: self.max_sessions]

        self._doc_keys: List[Tuple[str, int]] = all_doc_keys

        # Effective per-task limits based on selected doc_keys (used by evaluator).
        included_doc_ids_by_task: Dict[str, set[int]] = defaultdict(set)
        for tname, doc_id in self._doc_keys:
            included_doc_ids_by_task[tname].add(doc_id)

        self.limits: List[Optional[int]] = []
        for task_output in self.eval_tasks:
            tname = str(task_output.task_name)
            # Because we select a prefix via round-robin over per-task prefixes,
            # the number of docs included for a task is a valid `doc_iterator(limit=...)`.
            self.limits.append(len(included_doc_ids_by_task.get(tname, set())))

        # request_id -> instance/context length mapping for evaluator
        self.instance_by_request_id: Dict[int, Instance] = {}
        self.ctxlen_tokens_by_request_id: Dict[int, int] = {}

        self._cursor = 0
        self._next_request_id = 0
        self._instances_by_task_doc = task_to_instances_by_doc

    def _adjust_config(self, task_dict: dict) -> dict:
        adjusted_task_dict: dict = {}
        for task_name, task_obj in task_dict.items():
            if isinstance(task_obj, dict):
                adjusted_task_dict = {
                    **adjusted_task_dict,
                    **{task_name: self._adjust_config(task_obj)},
                }
                continue

            # override tasks' fewshot values unless task config forces 0
            default_num_fewshot = task_obj.get_config("num_fewshot")
            if default_num_fewshot == 0:
                logger.info(
                    "num_fewshot is 0 for %s in task config; leaving unchanged.",
                    task_name,
                )
            else:
                task_obj.set_config(key="num_fewshot", value=self.config.num_fewshot)

            task_obj.set_fewshot_seed(seed=self._fewshot_rng.randint(0, 2**32 - 1))
            adjusted_task_dict[task_name] = task_obj

        return adjusted_task_dict

    def capacity(self) -> int:
        return len(self._doc_keys)

    def generate_session(self) -> Session:
        if self._cursor >= len(self._doc_keys):
            raise StopIteration

        session_id = self._cursor
        task_name, doc_id = self._doc_keys[self._cursor]
        instances = self._instances_by_task_doc.get(task_name, {}).get(doc_id, [])

        session_graph = SessionGraph()
        requests: Dict[int, Request] = {}

        node_id = 0
        for instance in instances:
            repeats = int(getattr(instance, "repeats", 1) or 1)
            for _ in range(repeats):
                request_id = self._next_request_id
                self._next_request_id += 1

                add_node(session_graph, SessionNode(id=node_id, wait_after_ready=0.0))
                request = self._build_request(
                    request_id=request_id, instance=instance, node_id=node_id
                )
                requests[node_id] = request
                self.instance_by_request_id[request_id] = instance
                node_id += 1

        self._cursor += 1
        return Session(id=session_id, session_graph=session_graph, requests=requests)

    def _build_request(
        self, request_id: int, instance: Instance, node_id: int = 0
    ) -> Request:
        session_context = {
            "node_id": node_id,
            "wait_after_ready": 0.0,
            "parent_nodes": [],
            "history_parent": None,
        }

        channels: Dict[ChannelModality, Any] = {}
        metadata: Dict[str, Any] = {
            "lmeval_request_type": instance.request_type,
        }

        if instance.request_type == str(LMEvalOutputType.GENERATE_UNTIL):
            context, gen_kwargs, *rest = instance.args  # type: ignore
            multimodal_arg = rest[0] if rest and isinstance(rest[0], dict) else {}

            ctxlen = len(self.text_tokenizer.encode(context))
            max_gen_toks = gen_kwargs.get("max_gen_toks")

            channels[ChannelModality.TEXT] = TextChannelRequestContent(
                input_text=context,
                target_output_tokens=(
                    int(max_gen_toks) if isinstance(max_gen_toks, int) else 0
                ),
                target_prompt_tokens=ctxlen,
            )

            metadata["api_mode"] = "chat"
            metadata["sampling_params"] = dict(gen_kwargs)

        elif instance.request_type == str(LMEvalOutputType.LOGLIKELIHOOD):
            context, target, *rest = instance.args  # type: ignore
            multimodal_arg = rest[0] if rest and isinstance(rest[0], dict) else {}

            # Match lm-eval's token-boundary handling:
            # if `context` ends with spaces, move those spaces to the beginning of
            # `target` so the context/continuation boundary tokenizes correctly
            n_spaces = len(context) - len(context.rstrip())
            if n_spaces > 0:
                target = context[-n_spaces:] + target
                context = context[:-n_spaces]

            ctxlen = len(self.text_tokenizer.encode(context))
            prompt = context + target
            prompt_len = len(self.text_tokenizer.encode(prompt))

            channels[ChannelModality.TEXT] = TextChannelRequestContent(
                input_text=prompt,
                target_output_tokens=1,
                target_prompt_tokens=prompt_len,
            )
            if multimodal_arg:
                logger.warning(
                    "Ignoring multimodal inputs for loglikelihood request_id=%d",
                    request_id,
                )

            self.ctxlen_tokens_by_request_id[request_id] = ctxlen

            metadata["api_mode"] = "completions"
            metadata["sampling_params"] = {
                "stream": False,
                # Required for loglikelihood scoring: return logprobs for prompt tokens.
                # This matches upstream lm-eval's OpenAI completions behavior.
                "echo": True,
                # OpenAI-compatible completions API expects an integer `logprobs`
                # (how many top tokens to return). `1` is sufficient for
                # loglikelihood scoring and matches upstream lm-eval behavior.
                "logprobs": 1,
                "temperature": 0,
                # `/completions` uses `max_tokens` (not chat's
                # `max_completion_tokens`)
                "max_tokens": 1,
            }

        else:
            raise NotImplementedError(
                f"lm-eval request_type '{instance.request_type}' not supported yet."
            )

        return Request(
            id=request_id,
            channels=channels,
            session_context=session_context,
            metadata=metadata,
        )



import json

import pytest

from veeksha.config.evaluator import PerformanceEvaluatorConfig
from veeksha.core.request_content import TextChannelRequestContent
from veeksha.core.response import ChannelResponse, RequestResult
from veeksha.evaluator.performance.text import TextPerformanceEvaluator
from veeksha.types import ChannelModality


@pytest.mark.unit
def test_text_prefill_stats_groups_by_target_prompt_tokens(tmp_path) -> None:
    cfg = PerformanceEvaluatorConfig(stream_metrics=False)
    evaluator = TextPerformanceEvaluator(
        config=cfg,
        channel_config=cfg.text_channel,
        benchmark_start_time=0.0,
    )

    # Two prompt lengths (8 and 16), two samples each.
    prompts = [(0, 8, 0.10), (1, 8, 0.20), (2, 16, 0.30), (3, 16, 0.40)]
    for request_id, target_prompt_tokens, ttfc in prompts:
        evaluator.register_request(
            request_id=request_id,
            session_id=0,
            dispatched_at=0.0,
            content=TextChannelRequestContent(
                input_text="hello",
                target_output_tokens=2,
                target_prompt_tokens=target_prompt_tokens,
            ),
        )

        result = RequestResult(
            request_id=request_id,
            session_id=0,
            session_total_requests=1,
            channels={
                ChannelModality.TEXT: ChannelResponse(
                    modality=ChannelModality.TEXT,
                    content="xy",
                    metrics={
                        "is_stream": True,
                        "inter_chunk_times": [ttfc, 0.01],
                        "num_delta_prompt_tokens": target_prompt_tokens,
                        "num_total_prompt_tokens": target_prompt_tokens,
                        "num_output_tokens": 2,
                    },
                )
            },
            success=True,
            client_completed_at=ttfc + 0.01,
            client_picked_up_at=0.0,
            scheduler_ready_at=0.0,
            scheduler_dispatched_at=0.0,
            result_processed_at=ttfc + 0.01,
        )

        evaluator.record_request_completed(
            request_id=request_id,
            session_id=0,
            completed_at=ttfc + 0.01,
            response=result,
        )

    evaluator._save_prefill_stats(str(tmp_path))
    out_path = tmp_path / "prefill_stats.json"
    assert out_path.exists()

    data = json.loads(out_path.read_text())
    assert data["metric"] == "ttfc"
    assert data["group_by"] == "target_num_delta_prompt_tokens"

    assert data["groups"]["8"]["count"] == 2
    assert data["groups"]["8"]["mean"] == pytest.approx(0.15)
    assert data["groups"]["16"]["count"] == 2
    assert data["groups"]["16"]["mean"] == pytest.approx(0.35)






"""
Server configuration for LLM inference systems.
"""

from dataclasses import field
from typing import Any, Dict, List, Optional, Union

from veeksha.config.core.base_poly_config import BasePolyConfig
from veeksha.config.core.frozen_dataclass import frozen_dataclass
from veeksha.types import ServerType


@frozen_dataclass(allow_from_file=True)
class BaseServerConfig(BasePolyConfig):
    """Base configuration for server launch and management."""

    env_path: Optional[str] = field(
        default=None,
        metadata={"help": "Path to a Python environment directory (virtualenv/conda)."},
    )

    model: str = field(
        default="meta-llama/Meta-Llama-3-8B-Instruct",
        metadata={"help": "Model name or path."},
    )

    host: str = field(
        default="localhost", metadata={"help": "Host address for the server"}
    )

    port: int = field(default=8000, metadata={"help": "Port number for the server"})

    api_key: str = field(
        default="token-abc123", metadata={"help": "API key for server authentication"}
    )

    gpu_ids: Optional[List[int]] = field(
        default=None,
        metadata={"help": "List of GPU IDs to use (None means auto-assign)"},
    )

    startup_timeout: int = field(
        default=300, metadata={"help": "Timeout in seconds for server startup"}
    )

    health_check_interval: float = field(
        default=2.0, metadata={"help": "Interval in seconds between health checks"}
    )

    require_contiguous_gpus: bool = field(
        default=True,
        metadata={
            "help": "Require contiguous GPU allocation (e.g., GPUs 0,1,2 instead of 0,2,5)"
        },
    )

    # engine arguments

    tensor_parallel_size: int = field(
        default=1, metadata={"help": "Number of GPUs for tensor parallelism"}
    )

    dtype: str = field(
        default="auto",
        metadata={
            "help": "Data type for model weights (auto, float16, bfloat16, etc.)"
        },
    )

    max_model_len: Optional[int] = field(
        default=None, metadata={"help": "Maximum model context length"}
    )

    additional_args: Union[str, Dict[str, Any], None] = field(
        default="{}",
        metadata={
            "help": "Additional engine-specific arguments as JSON string, dict, or None."
        },
    )

    def get_api_base_url(self) -> str:
        return f"http://{self.host}:{self.port}/v1"

    def get_health_check_url(self) -> str:
        return f"http://{self.host}:{self.port}/health"

    def get_gpu_env_var(self) -> Optional[str]:
        """Get CUDA_VISIBLE_DEVICES value if gpu_ids is specified."""
        if self.gpu_ids is not None:
            return ",".join(map(str, self.gpu_ids))
        return None

    def get_num_gpus(self) -> int:
        """Get the number of GPUs required for this server."""
        if self.gpu_ids is not None:
            return len(self.gpu_ids)
        return self.tensor_parallel_size

    @property
    def engine(self) -> str:
        """Get the engine name for logging/compat."""
        return self.get_type().name.lower()


@frozen_dataclass
class VllmServerConfig(BaseServerConfig):
    @classmethod
    def get_type(cls) -> ServerType:
        return ServerType.VLLM


@frozen_dataclass
class VajraServerConfig(BaseServerConfig):
    @classmethod
    def get_type(cls) -> ServerType:
        return ServerType.VAJRA


@frozen_dataclass
class SglangServerConfig(BaseServerConfig):
    @classmethod
    def get_type(cls) -> ServerType:
        return ServerType.SGLANG



import pytest

from veeksha.config.generator.length import StairLengthGeneratorConfig
from veeksha.generator.length.stair import StairLengthGenerator


@pytest.mark.unit
def test_stair_length_generator_repeats_and_wraps() -> None:
    cfg = StairLengthGeneratorConfig(values=[1, 2, 3], repeat_each=2, wrap=True)
    gen = StairLengthGenerator(cfg)

    seq = [gen.get_next_value() for _ in range(8)]
    assert seq == [1, 1, 2, 2, 3, 3, 1, 1]


@pytest.mark.unit
def test_stair_length_generator_can_hold_last_value_without_wrap() -> None:
    cfg = StairLengthGeneratorConfig(values=[5, 7], repeat_each=1, wrap=False)
    gen = StairLengthGenerator(cfg)

    seq = [gen.get_next_value() for _ in range(5)]
    assert seq == [5, 7, 7, 7, 7]








"""Capacity search runs a benchmark multiple times with the same configuration,
varying only a single traffic scheduler knob. It then finds the maximum value
that still satisfies all configured SLOs, as recorded in `metrics/slo_results.json`.
"""

from __future__ import annotations

import hashlib
import json
import os
import shutil
from dataclasses import replace
from datetime import datetime, timezone
from typing import Any, Callable, Dict, Optional, Tuple, cast

import yaml

from veeksha.benchmark import manage_benchmark_run
from veeksha.config.benchmark import BenchmarkConfig
from veeksha.config.capacity_search import CapacitySearchConfig
from veeksha.config.generator.interval import (
    BaseIntervalGeneratorConfig,
    FixedIntervalGeneratorConfig,
)
from veeksha.config.traffic import ConcurrentTrafficConfig, RateTrafficConfig
from veeksha.config.utils import dataclass_to_dict
from veeksha.logger import init_logger
from veeksha.wandb_integration import (
    dedup_tags,
    maybe_log_capacity_search_summary,
    update_run_tags,
)

logger = init_logger(__name__)


def _persist_capacity_search_config_yaml(config: CapacitySearchConfig) -> str:
    """Write the resolved capacity search configuration to config.yml."""
    os.makedirs(config.output_dir, exist_ok=True)
    config_dict = dataclass_to_dict(config)
    config_path = os.path.join(config.output_dir, "config.yml")
    with open(config_path, "w", encoding="utf-8") as config_file:
        yaml.safe_dump(
            config_dict,
            config_file,
            sort_keys=False,
            default_flow_style=False,
            allow_unicode=True,
        )
    return config_path


def _init_capacity_search_output_dir(config: CapacitySearchConfig) -> str:
    """Resolve and prepare the final capacity search output directory.

    Mirrors `veeksha.new.benchmark_utils._init_output_dir` semantics:
    - persist config to YAML
    - hash the persisted config
    - create a timestamp+hash subdir inside the user-specified base output dir
    - move config.yml into the resolved dir
    - update `config.output_dir` in-place
    """
    base_output_dir = config.output_dir
    os.makedirs(base_output_dir, exist_ok=True)

    config_path = _persist_capacity_search_config_yaml(config)
    with open(config_path, "rb") as config_file:
        config_bytes = config_file.read()
    config_hash = hashlib.sha1(config_bytes).hexdigest()[:8]

    timestamp_prefix = datetime.now(timezone.utc).strftime("%d:%m:%Y-%H:%M:%S")
    base_dir_name = f"{timestamp_prefix}-{config_hash}"
    resolved_output_dir = os.path.join(base_output_dir, base_dir_name)

    suffix = 1
    while os.path.exists(resolved_output_dir):
        suffix += 1
        resolved_output_dir = os.path.join(base_output_dir, f"{base_dir_name}-{suffix}")

    os.makedirs(resolved_output_dir, exist_ok=True)
    shutil.move(config_path, os.path.join(resolved_output_dir, "config.yml"))
    object.__setattr__(config, "output_dir", resolved_output_dir)
    logger.info("Capacity search outputs will be stored in %s", resolved_output_dir)
    return resolved_output_dir


def _read_slo_results(run_output_dir: str) -> Dict[str, Any]:
    path = os.path.join(run_output_dir, "metrics", "slo_results.json")
    if not os.path.exists(path):
        raise FileNotFoundError(
            f"Missing SLO results at '{path}'. Ensure the benchmark evaluator writes "
            "request-level metrics and has at least one configured SLO."
        )
    with open(path, "r", encoding="utf-8") as f:
        data = json.load(f)
    if not isinstance(data, dict):
        raise ValueError(f"Invalid slo_results.json at '{path}': expected dict.")
    return cast(Dict[str, Any], data)


def _knob_description(benchmark_config: BenchmarkConfig) -> Tuple[str, str]:
    traffic = benchmark_config.traffic_scheduler
    if isinstance(traffic, RateTrafficConfig):
        return ("rate", "traffic_scheduler.interval_generator.arrival_rate")
    if isinstance(traffic, ConcurrentTrafficConfig):
        return ("concurrent", "traffic_scheduler.target_concurrent_sessions")
    raise ValueError(
        f"Unsupported traffic scheduler type: {type(traffic).__name__}. "
        "Supported: rate, concurrent."
    )


def patch_traffic_knob(
    benchmark_config: BenchmarkConfig, *, value: float
) -> BenchmarkConfig:
    """Return a copy of BenchmarkConfig with traffic knob set to the provided value."""
    traffic = benchmark_config.traffic_scheduler

    if isinstance(traffic, RateTrafficConfig):
        interval_cfg: BaseIntervalGeneratorConfig = traffic.interval_generator
        if isinstance(interval_cfg, FixedIntervalGeneratorConfig):
            # `value` as sessions-per-second and convert to fixed interval.
            new_interval_cfg = replace(interval_cfg, interval=float(1.0 / value))
        else:
            new_interval_cfg = replace(
                cast(Any, interval_cfg), arrival_rate=float(value)
            )
        new_traffic = replace(traffic, interval_generator=new_interval_cfg)
        return replace(benchmark_config, traffic_scheduler=new_traffic)

    if isinstance(traffic, ConcurrentTrafficConfig):
        if abs(value - round(value)) > 1e-9:
            raise ValueError(
                f"Concurrent capacity search requires integer values; got {value}."
            )
        target = int(round(value))
        # concurrent traffic ramps at ~1 session/sec
        new_traffic = replace(
            traffic,
            target_concurrent_sessions=target,
            rampup_seconds=target,
        )
        return replace(benchmark_config, traffic_scheduler=new_traffic)

    raise ValueError(
        f"Unsupported traffic scheduler type: {type(traffic).__name__}. "
        "Supported: rate, concurrent."
    )


def _adaptive_capacity_search(
    *,
    start_value: float,
    max_value: float,
    expansion_factor: float,
    is_passing: Callable[[float], bool],
    max_iterations: int,
    precision: int,
    integer_domain: bool,
) -> Tuple[Optional[float], int]:
    """Adaptive two-phase search: exponential probe then binary search."""
    if max_iterations <= 0:
        raise ValueError("max_iterations must be > 0")
    if start_value <= 0:
        raise ValueError("start_value must be > 0")
    if expansion_factor <= 1.0:
        raise ValueError("expansion_factor must be > 1.0")
    if start_value > max_value:
        raise ValueError("start_value must be <= max_value")

    def _round_value(v: float) -> float:
        if integer_domain:
            return float(max(1, int(round(v))))
        scale = 10**precision
        return round(v * scale) / scale

    iters = 0
    probe = _round_value(start_value)
    last_passing: Optional[float] = None
    first_failing: Optional[float] = None

    # exponential probing
    logger.info("Phase 1: Probing from %s (expansion=%s)", probe, expansion_factor)

    while iters < max_iterations:
        iters += 1
        if is_passing(probe):
            last_passing = probe
            if probe >= max_value:
                logger.info("Passed at max_value=%s", max_value)
                return (last_passing, iters)
            next_probe = _round_value(probe * expansion_factor)
            if next_probe <= probe:
                next_probe = (
                    probe + 1 if integer_domain else probe + (1.0 / 10**precision)
                )
            probe = min(_round_value(next_probe), _round_value(max_value))
        else:
            first_failing = probe
            break

    if last_passing is None:
        logger.warning("Start value %s failed", start_value)
        return (None, iters)

    if first_failing is None:
        return (last_passing, iters)

    # binary search
    logger.info("Phase 2: Binary search in [%s, %s]", last_passing, first_failing)

    if integer_domain:
        lo = int(round(last_passing))
        hi = int(round(first_failing))
        best = lo

        while lo < hi and iters < max_iterations:
            mid = (lo + hi + 1) // 2
            iters += 1
            if is_passing(float(mid)):
                best = mid
                lo = mid
            else:
                hi = mid - 1

        return (float(best), iters)

    else:
        scale = 10**precision
        lo = int(round(last_passing * scale))
        hi = int(round(first_failing * scale))
        best_i = lo

        while lo < hi and iters < max_iterations:
            mid_i = (lo + hi + 1) // 2
            mid = mid_i / scale
            iters += 1
            if is_passing(mid):
                best_i = mid_i
                lo = mid_i
            else:
                hi = mid_i - 1

        return (best_i / scale, iters)


def run_capacity_search(config: CapacitySearchConfig) -> Dict[str, Any]:
    """Run capacity search and return a machine-readable result dict."""
    _init_capacity_search_output_dir(config)
    runs_dir = os.path.join(config.output_dir, "runs")
    os.makedirs(runs_dir, exist_ok=True)
    base_benchmark_config = config.benchmark_config
    scheduler_type, knob_path = _knob_description(base_benchmark_config)

    os.makedirs(config.output_dir, exist_ok=True)

    iterations: list[Dict[str, Any]] = []
    attempt_counter = 0

    logger.info(
        "Capacity search: knob=%s | start=%s, max=%s, expansion=%s | max_iterations=%s | output_dir=%s",
        knob_path,
        config.start_value,
        config.max_value,
        config.expansion_factor,
        config.max_iterations,
        config.output_dir,
    )

    def run_one(value: float) -> bool:
        nonlocal attempt_counter
        attempt_counter += 1
        separator = "-" * 88
        logger.info(separator)
        logger.info(
            "Capacity search attempt %d | %s=%s",
            attempt_counter,
            knob_path,
            value,
        )
        # ensure run outputs are grouped under capacity search output_dir
        run_cfg = replace(base_benchmark_config, output_dir=runs_dir)
        run_cfg = patch_traffic_knob(run_cfg, value=value)

        # Group all attempts under one wandb group and give them readable names.
        if getattr(run_cfg, "wandb", None) and run_cfg.wandb.enabled:
            auto_group = f"capsearch-{os.path.basename(config.output_dir.rstrip('/'))}"
            short_knob = knob_path.rsplit(".", 1)[-1]
            run_cfg = replace(
                run_cfg,
                wandb=replace(
                    run_cfg.wandb,
                    group=run_cfg.wandb.group or auto_group,
                    run_name=run_cfg.wandb.run_name
                    or f"{attempt_counter:02d}-{short_knob}={value}",
                    tags=dedup_tags([*run_cfg.wandb.tags, "capsearch"]),
                ),
            )
        manage_benchmark_run(run_cfg)

        slo = _read_slo_results(run_cfg.output_dir)
        all_met = bool(slo.get("all_slos_met", False))

        slo_summaries = []
        for r in slo.get("results", []) if isinstance(slo.get("results"), list) else []:
            if not isinstance(r, dict):
                continue
            name = r.get("name") or r.get("slo_metric_key") or "slo"
            met = r.get("met")
            observed = r.get("observed_value")
            threshold = r.get("threshold")
            if observed is None or threshold is None:
                slo_summaries.append(f"{name}={met}")
            else:
                slo_summaries.append(f"{name}:{observed:.4f}<={threshold:.4f}={met}")

        logger.info(
            "Capacity search attempt %d result: %s | run_dir=%s%s",
            attempt_counter,
            "PASS" if all_met else "FAIL",
            run_cfg.output_dir,
            (f" | {' | '.join(slo_summaries)}" if slo_summaries else ""),
        )
        iterations.append(
            {
                "value": value,
                "all_slos_met": all_met,
                "run_dir": run_cfg.output_dir,
                "slo_results": slo,
            }
        )
        return all_met

    integer_domain = scheduler_type == "concurrent"

    best_value, iters_used = _adaptive_capacity_search(
        start_value=config.start_value,
        max_value=config.max_value,
        expansion_factor=config.expansion_factor,
        is_passing=run_one,
        max_iterations=config.max_iterations,
        precision=config.precision,
        integer_domain=integer_domain,
    )

    # If we hit the search budget, we may not have proven the true maximum.
    if (
        iters_used >= config.max_iterations
        and best_value is not None
        and best_value != config.max_value
    ):
        logger.warning(
            "Capacity search stopped after max_iterations=%s without reaching max_value=%s; "
            "current best=%s. Increase max_iterations to fully search the range.",
            config.max_iterations,
            config.max_value,
            best_value,
        )

    best_run_dir = None
    if best_value is not None:
        # last passing run for the best value
        for entry in reversed(iterations):
            if entry["value"] == best_value and entry["all_slos_met"]:
                best_run_dir = entry["run_dir"]
                break

    result: Dict[str, Any] = {
        "traffic_scheduler_type": scheduler_type,
        "searched_knob": knob_path,
        "start_value": config.start_value,
        "max_value": config.max_value,
        "expansion_factor": config.expansion_factor,
        "max_iterations": config.max_iterations,
        "precision": config.precision,
        "iterations_run": iters_used,
        "best_value": best_value,
        "best_run_dir": best_run_dir,
        "history": iterations,
    }

    out_path = os.path.join(config.output_dir, "capacity_search_results.json")
    with open(out_path, "w", encoding="utf-8") as f:
        json.dump(result, f, indent=2)
    logger.info("Capacity search result written to %s", out_path)
    if best_value is None:
        logger.warning(
            "Capacity search complete: no passing value found starting from %s for %s",
            config.start_value,
            knob_path,
        )
    else:
        logger.info(
            "Capacity search best: %s=%s | run_dir=%s",
            knob_path,
            best_value,
            best_run_dir,
        )
        if best_run_dir:
            update_run_tags(best_run_dir, ["BEST_CONFIG"])

    bench_cfg = config.benchmark_config
    if getattr(bench_cfg, "wandb", None) and bench_cfg.wandb.enabled:
        auto_group = f"capsearch-{os.path.basename(config.output_dir.rstrip('/'))}"
        maybe_log_capacity_search_summary(
            output_dir=config.output_dir,
            wandb_cfg=bench_cfg.wandb,
            result=result,
            group=bench_cfg.wandb.group or auto_group,
        )
    return result


def main() -> None:
    configs = CapacitySearchConfig.create_from_cli_args()
    for cfg in configs:
        logger.info("Running capacity search (output_dir=%s)", cfg.output_dir)
        run_capacity_search(cfg)


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from veeksha.logger import init_logger
from veeksha.slo.metrics import lower_is_better
from veeksha.slo.slo import SloSet

logger = init_logger(__name__)


@dataclass(frozen=True)
class SloCheckResult:
    """Result for a single evaluated SLO."""

    met: bool
    slo_metric_key: str
    observed_value: float
    threshold: float
    percentile: float
    metric: Optional[str]
    name: Optional[str]
    lower_is_better: bool

    def to_dict(self) -> Dict[str, Any]:
        return {
            "met": self.met,
            "slo_metric_key": self.slo_metric_key,
            "observed_value": self.observed_value,
            "threshold": self.threshold,
            "percentile": self.percentile,
            "metric": self.metric,
            "name": self.name,
            "lower_is_better": self.lower_is_better,
        }


@dataclass(frozen=True)
class SloEvaluationResult:
    """Aggregate SLO evaluation result for a benchmark run."""

    all_slos_met: bool
    results: List[SloCheckResult]

    def to_dict(self) -> Dict[str, Any]:
        return {
            "all_slos_met": self.all_slos_met,
            "results": [r.to_dict() for r in self.results],
        }


class SloEvaluator:
    """Evaluates a SloSet against request-level metrics."""

    def __init__(self, slo_set: SloSet):
        self.slo_set = slo_set

    def evaluate(self, request_level_metrics: Dict[str, Any]) -> SloEvaluationResult:
        """Evaluate all SLOs.

        Returns:
            SloEvaluationResult with per-SLO details.
        """

        results: List[SloCheckResult] = []

        for slo in self.slo_set.slos:
            met, observed = slo.evaluate(request_level_metrics)
            metric_key = slo.get_slo_metric_key()

            threshold = slo.get_threshold()
            percentile = getattr(slo.config, "percentile", 1.0)
            metric = getattr(slo.config, "metric", None)
            name = getattr(slo.config, "name", None)

            lib = lower_is_better(metric) if isinstance(metric, str) else True

            results.append(
                SloCheckResult(
                    met=met,
                    slo_metric_key=metric_key,
                    observed_value=float(observed),
                    threshold=float(threshold),
                    percentile=float(percentile),
                    metric=metric,
                    name=name,
                    lower_is_better=lib,
                )
            )

            logger.info(
                "SLO '%s' %s (value=%.4f threshold=%.4f)",
                name or str(slo),
                "MET" if met else "MISSED",
                observed,
                threshold,
            )

        all_met = all(r.met for r in results) if results else True
        logger.info("All SLOs met: %s", all_met)
        return SloEvaluationResult(all_slos_met=all_met, results=results)



#!/usr/bin/env python3
"""Minimal OpenAI-compatible server used for orchestration E2E tests."""

from __future__ import annotations

import argparse
import json
import sys
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from typing import Dict, Tuple


class _RequestHandler(BaseHTTPRequestHandler):
    server_version = "DummyLLM/0.1"

    def _send_json(self, payload: Dict, status: HTTPStatus = HTTPStatus.OK) -> None:
        body = json.dumps(payload).encode("utf-8")
        self.send_response(status)
        self.send_header("Content-Type", "application/json")
        self.send_header("Content-Length", str(len(body)))
        self.end_headers()
        self.wfile.write(body)

    def do_GET(self) -> None:  # noqa: N802 (BaseHTTPRequestHandler API)
        if self.path == "/health":
            self._send_json({"status": "ok"})
        elif self.path == "/v1/models":
            model_name = getattr(self.server, "model_name", "dummy")
            payload = {"data": [{"id": model_name, "object": "model"}]}
            self._send_json(payload)
        else:
            self.send_error(HTTPStatus.NOT_FOUND, "Unknown path")

    def do_POST(self) -> None:  # noqa: N802 (BaseHTTPRequestHandler API)
        if self.path.endswith("/chat/completions"):
            length = int(self.headers.get("Content-Length", "0"))
            raw_body = self.rfile.read(length) if length > 0 else b"{}"
            try:
                body = json.loads(raw_body.decode("utf-8"))
            except json.JSONDecodeError:
                self.send_error(HTTPStatus.BAD_REQUEST, "Invalid JSON")
                return

            messages = body.get("messages", [])
            reply_text = "Hello from dummy server"
            if messages:
                reply_text = f"Echo: {messages[-1].get('content', '')}"

            response = {
                "id": "dummy-chatcmpl",
                "object": "chat.completion",
                "choices": [
                    {
                        "index": 0,
                        "finish_reason": "stop",
                        "message": {"role": "assistant", "content": reply_text},
                    }
                ],
            }
            self._send_json(response)
        else:
            self.send_error(HTTPStatus.NOT_FOUND, "Unknown path")

    def log_message(self, format: str, *args) -> None:  # noqa: A003
        # Silence noisy default logging during tests
        return


def _parse_args() -> Tuple[str, int, str]:
    parser = argparse.ArgumentParser(description="Run a dummy LLM server")
    parser.add_argument("--host", default="127.0.0.1", help="Host to bind")
    parser.add_argument("--port", type=int, required=True, help="Port to bind")
    parser.add_argument("--model", default="dummy", help="Model name")
    args = parser.parse_args()
    return args.host, args.port, args.model


def main() -> None:
    host, port, model = _parse_args()
    server = ThreadingHTTPServer((host, port), _RequestHandler)
    server.model_name = model  # type: ignore[attr-defined]

    # NOTE: We intentionally do not register SIGTERM/SIGINT handlers. Calling
    # server.shutdown() from a signal handler can deadlock if shutdown waits
    # for serve_forever() to exit on the same thread. We rely on the
    # parent/test harness to terminate this process and the existing
    # KeyboardInterrupt handling below to close the server gracefully.

    try:
        server.serve_forever()
    except KeyboardInterrupt:
        pass
    finally:
        server.server_close()


if __name__ == "__main__":
    try:
        main()
    except Exception as exc:  # pragma: no cover - surfaced in test logs
        print(f"Dummy server failed: {exc}", file=sys.stderr)
        raise



from __future__ import annotations

import hashlib
import itertools
import random
from typing import Callable, TypeVar

import numpy as np

T = TypeVar("T")

_MAX_SEED = 2**32


def derive_seed(root_seed: int, *path: str) -> int:
    """Derive a deterministic child seed from a root seed and path components."""
    hasher = hashlib.blake2b(digest_size=8)
    hasher.update(root_seed.to_bytes(8, byteorder="big", signed=False))
    for component in path:
        hasher.update(component.encode("utf-8"))
    digest = int.from_bytes(hasher.digest(), byteorder="big", signed=False)
    return digest % _MAX_SEED


def random_from_seed(seed: int) -> random.Random:
    return random.Random(seed)


def numpy_random_from_seed(seed: int) -> np.random.RandomState:
    return np.random.RandomState(seed)


def _make_factory(
    root_seed: int,
    initializer: Callable[[int], T],
    path: tuple[str, ...],
) -> Callable[[], T]:
    counter = itertools.count()

    def factory() -> T:
        seed = derive_seed(root_seed, *path, str(next(counter)))
        return initializer(seed)

    return factory


def numpy_factory(root_seed: int, *path: str) -> Callable[[], np.random.RandomState]:
    # when something requires a np.random.RandomState, we use this function
    return _make_factory(root_seed, numpy_random_from_seed, path)


def random_factory(root_seed: int, *path: str) -> Callable[[], random.Random]:
    # when something requires a random.Random, we use this function
    return _make_factory(root_seed, random_from_seed, path)


def random_for_path(root_seed: int, *path: str) -> random.Random:
    return random_from_seed(derive_seed(root_seed, *path))


class SeedManager:
    """Hierarchical seed manager yielding deterministic RNG streams."""

    def __init__(self, seed: int):
        self._seed = seed % _MAX_SEED

    def child(self, *path: str) -> "SeedManager":
        return SeedManager(derive_seed(self._seed, *path))

    def random_factory(self, *path: str) -> Callable[[], random.Random]:
        return random_factory(self._seed, *path)

    def numpy_factory(self, *path: str) -> Callable[[], np.random.RandomState]:
        return numpy_factory(self._seed, *path)

    def random(self, *path: str) -> random.Random:
        return random_for_path(self._seed, *path)

    def numpy(self, *path: str) -> np.random.RandomState:
        return numpy_random_from_seed(derive_seed(self._seed, *path))

    @property
    def seed(self) -> int:
        return self._seed



from dataclasses import field

from veeksha.config.core.base_poly_config import BasePolyConfig
from veeksha.config.core.frozen_dataclass import frozen_dataclass
from veeksha.config.generator.interval import (
    BaseIntervalGeneratorConfig,
    PoissonIntervalGeneratorConfig,
)
from veeksha.types import TrafficType


@frozen_dataclass(allow_from_file=True)
class BaseTrafficConfig(BasePolyConfig):
    cancel_session_on_failure: bool = field(
        default=True,
        metadata={"help": "Whether to cancel the session on failure of any request."},
    )


@frozen_dataclass
class RateTrafficConfig(BaseTrafficConfig):
    interval_generator: BaseIntervalGeneratorConfig = field(
        default_factory=PoissonIntervalGeneratorConfig,
        metadata={
            "help": "Interval generator for the traffic (sessions per second). Available: poisson, gamma, fixed."
        },
    )

    @classmethod
    def get_type(cls) -> TrafficType:
        return TrafficType.RATE


@frozen_dataclass
class ConcurrentTrafficConfig(BaseTrafficConfig):
    target_concurrent_sessions: int = field(
        default=3,
        metadata={"help": "Target number of concurrent sessions to maintain."},
    )
    rampup_seconds: int = field(
        default=10,
        metadata={
            "help": "Number of seconds to ramp up the traffic. i.e. 'Take 10 seconds to ramp up to the target concurrent sessions.'"
        },
    )

    @classmethod
    def get_type(cls) -> TrafficType:
        return TrafficType.CONCURRENT



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Set

from veeksha.config.evaluator import BaseEvaluatorConfig
from veeksha.core.seeding import SeedManager
from veeksha.types import ChannelModality


@dataclass
class EvaluationResult:
    """Container for evaluation results from any evaluator.

    Attributes:
        evaluator_type: String identifying the evaluator type (e.g., "performance", "accuracy")
        channel: Optional channel modality this result applies to (None for aggregate results)
        metrics: Dictionary of computed metrics
        raw_data: Optional dictionary containing detailed per-request data
    """

    evaluator_type: str
    channel: Optional[ChannelModality]
    metrics: Dict[str, Any]
    raw_data: Optional[Dict[str, Any]] = field(default=None)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "evaluator_type": self.evaluator_type,
            "channel": str(self.channel) if self.channel else None,
            "metrics": self.metrics,
        }


class BaseEvaluator(ABC):
    """Abstract base class for all evaluators.

    Evaluators consume completed requests/responses and produce evaluation metrics.
    They might operate on specific channels or aggregate across all channels.

    The lifecycle would be:
    1. Construction with config
    2. register_request() called for each request when dispatched
    3. record_request_completed() called when response received
    4. record_session_completed() called when session finishes
    5. finalize() called at the end to compute aggregate metrics
    6. save() called to persist results to disk
    """

    def __init__(
        self,
        config: BaseEvaluatorConfig,
        seed_manager: Optional[SeedManager] = None,
    ):
        """Initialize the evaluator.

        Args:
            config: Evaluator configuration
            seed_manager: Optional seed manager for reproducibility
        """
        self.config = config
        self.seed_manager = seed_manager
        self._target_channels = (
            set(config.target_channels) if config.target_channels else None
        )

    def should_evaluate_channel(self, channel: ChannelModality) -> bool:
        """Check if this evaluator should process a given channel.

        Args:
            channel: The channel modality to check
        """
        if self._target_channels is None:
            return True  # all channels
        return channel in self._target_channels

    @abstractmethod
    def register_request(
        self,
        request_id: int,
        session_id: int,
        dispatched_at: float,
        channels: Dict[ChannelModality, Any],
    ) -> None:
        """Register a request that was dispatched to the server

        Args:
            request_id: Unique request id
            session_id: Session id this request belongs to
            dispatched_at: Monotonic timestamp when request was dispatched
            channels: Request content indexed by modality
        """
        raise NotImplementedError

    @abstractmethod
    def record_request_completed(
        self,
        request_id: int,
        session_id: int,
        completed_at: float,
        response: Any,
        error: Optional[Exception] = None,
    ) -> None:
        """Record that a request completed.

        Args:
            request_id: Unique request id
            session_id: Session id this request belongs to
            completed_at: Monotonic timestamp when response was received
            response: Response object containing timing and content data
            error: Optional exception if the request failed
        """
        raise NotImplementedError

    @abstractmethod
    def record_session_completed(
        self,
        session_id: int,
        completed_at: float,
        success: bool,
    ) -> None:
        """Record that a session completed.

        Args:
            session_id: Session id
            completed_at: Monotonic timestamp when session completed
            success: True if all requests in session succeeded
        """
        raise NotImplementedError

    @abstractmethod
    def finalize(self) -> EvaluationResult:
        """Finalize evaluation and return results.

        Returns:
            EvaluationResult containing computed metrics
        """
        raise NotImplementedError

    @abstractmethod
    def save(self, output_dir: str) -> None:
        """Save evaluation artifacts to the output directory.

        Args:
            output_dir: Directory to write output files to
        """
        raise NotImplementedError

    def get_streaming_metrics(self) -> Optional[Dict[str, Any]]:
        """Return current metrics for streaming updates.

        Returns:
            Dictionary of current metrics, or None if streaming not supported
        """
        return None

    def get_completed_request_count(self) -> int:
        """Return count of completed requests for progress tracking.

        Returns:
            Number of completed requests. Subclasses should override.
        """
        return 0

    def get_session_counts(self) -> tuple:
        """Return session counts for progress tracking.

        Returns:
            Tuple of (completed_sessions, errored_sessions, in_progress_sessions).
            Subclasses should override.
        """
        return (0, 0, 0)

    def set_included_requests(self, request_ids: Set[int]) -> None:
        """Set which requests to include in final metrics.

        If not called, all requests are included. Subclasses can override
        to filter metrics during finalization.

        Args:
            request_ids: Set of request IDs to include in metrics
        """



from veeksha.types.base_int_enum import BaseIntEnum


# ----- Traffic -----
class TrafficType(BaseIntEnum):
    RATE = 1
    CONCURRENT = 2


# ----- Interval / length generators -----
class IntervalGeneratorType(BaseIntEnum):
    POISSON = 1
    GAMMA = 2
    FIXED = 3


class LengthGeneratorType(BaseIntEnum):
    ZIPF = 1
    UNIFORM = 2
    FIXED = 3
    FIXED_STAIR = 4


# ----- Content -----
class SessionGeneratorType(BaseIntEnum):
    SYNTHETIC = 1
    LMEVAL = 2
    TRACE = 3


class TraceFlavorType(BaseIntEnum):
    CLAUDE_CODE = 1
    MOONCAKE_CONV = 2
    RAG = 3


class ChannelModality(BaseIntEnum):
    TEXT = 1
    IMAGE = 2
    AUDIO = 3
    VIDEO = 4


class SessionGraphType(BaseIntEnum):
    LINEAR = 1
    SINGLE_REQUEST = 2
    BRANCHING = 3


# ----- Evaluation -----
class EvaluationType(BaseIntEnum):
    PERFORMANCE = 1
    ACCURACY_LMEVAL = 2


# ----- Client -----
class ClientType(BaseIntEnum):
    OPENAI_CHAT_COMPLETIONS = 1
    OPENAI_COMPLETIONS = 2
    OPENAI_ROUTER = 3


# ----- Server -----
class ServerType(BaseIntEnum):
    VLLM = 1
    VAJRA = 2
    SGLANG = 3


# ----- SLOs -----
class SloType(BaseIntEnum):
    CONSTANT = 1


# ----- LMEval -----
class LMEvalOutputType(BaseIntEnum):
    LOGLIKELIHOOD = 1
    LOGLIKELIHOOD_ROLLING = 2
    GENERATE_UNTIL = 3
    MULTIPLE_CHOICE = 4



from abc import abstractmethod
from typing import Optional

from veeksha.config.generator.session_graph import BaseSessionGraphGeneratorConfig
from veeksha.core.seeding import SeedManager
from veeksha.core.session_graph import SessionGraph


class BaseSessionGraphGenerator:
    def __init__(
        self,
        config: BaseSessionGraphGeneratorConfig,
        seed_manager: Optional[SeedManager] = None,
    ):
        self.config = config
        self.seed_manager = seed_manager or SeedManager(0)

    @abstractmethod
    def generate_session_graph(self) -> SessionGraph:
        pass



import os
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Tuple, cast

import numpy as np
import pandas as pd

from veeksha.config.benchmark import BenchmarkConfig
from veeksha.logger import init_logger
from veeksha.types import IntervalGeneratorType, TrafficType

logger = init_logger(__name__)


@dataclass
class TestResult:
    summary: Dict[str, Any]
    passed: bool


class HealthChecker:
    def __init__(
        self,
        trace_file: str,
        metrics_file: str,
        benchmark_config: BenchmarkConfig,
    ):
        self.trace_file = trace_file
        self.metrics_file = metrics_file
        self.config = benchmark_config
        self.trace_df = pd.DataFrame()
        self.metrics_df = pd.DataFrame()
        self.merged_df = pd.DataFrame()

    def _get_col(self, col: str) -> str:
        """Get the actual column name after merge (handles suffixed columns)."""
        if col in self.merged_df.columns:
            return col
        # prefer trace over metrics
        if f"{col}_trace" in self.merged_df.columns:
            return f"{col}_trace"
        if f"{col}_metrics" in self.merged_df.columns:
            return f"{col}_metrics"
        return col

    def load_data(self) -> bool:
        try:
            self.metrics_df = pd.read_json(self.metrics_file, lines=True)
            if self.metrics_df.empty:
                logger.warning("Metrics file is empty.")
                return False

            try:
                if os.path.exists(self.trace_file):
                    self.trace_df = pd.read_json(self.trace_file, lines=True)
                else:
                    logger.warning(f"Trace file not found: {self.trace_file}")
                    self.trace_df = pd.DataFrame()
            except Exception as e:
                logger.warning(f"Failed to load trace file; proceeding without it: {e}")
                self.trace_df = pd.DataFrame()

            if not self.trace_df.empty:
                self.merged_df = pd.merge(
                    self.trace_df,
                    self.metrics_df,
                    on="request_id",
                    how="inner",
                    suffixes=("_trace", "_metrics"),
                )
            else:
                self.merged_df = self.metrics_df.copy()

            return True
        except Exception as e:
            logger.error(f"Failed to load health check data: {e}")
            return False

    def run_checks(self) -> List[TestResult]:
        if self.merged_df.empty:
            if not self.load_data():
                return []

        results = []

        # common checks
        results.append(self.check_intra_session_request_arrival())

        # scheduler-specific checks
        traffic_type = self.config.traffic_scheduler.get_type()
        if traffic_type == TrafficType.CONCURRENT:
            results.append(self.check_session_concurrency())
        elif traffic_type == TrafficType.RATE:
            results.append(self.check_session_dispatch_rate())

        # prompt length check
        results.append(self.check_prompt_length())

        # output length check
        results.append(self.check_output_length())

        # lifecycle timing delays check
        results.append(self.check_lifecycle_timing_delays())

        return results

    def run_and_save(self, output_path: str) -> List[TestResult]:
        """Run all health checks and save formatted results to a file.

        Args:
            output_path: Path to save the health check results.

        Returns:
            List of TestResult objects.
        """
        results = self.run_checks()

        with open(output_path, "w") as f:
            for result in results:
                f.write("=" * 60 + "\n")
                f.write(f"{result.summary['name'].upper()}\n")
                f.write("=" * 60 + "\n")
                f.write(f"Result: {'PASSED' if result.passed else 'FAILED'}\n\n")
                for section in result.summary["sections"]:
                    f.write(f"{section['title']}:\n")
                    for k, v in section["results"].items():
                        f.write(f"  {k:30s} {v}\n")
                    f.write("\n")
                f.write("\n")

        logger.info(f"Health checks completed. Results saved to {output_path}")
        return results

    def check_intra_session_request_arrival(self) -> TestResult:
        """
        Verify that requests are dispatched only after their dependencies are met
        plus the specified wait_after_ready time.

        Requires trace data for dependency graphs (session_context).
        """
        if self.trace_df.empty:
            trace_exists = os.path.exists(self.trace_file)
            reason = (
                f"Trace file '{self.trace_file}' not found."
                if not trace_exists
                else f"Trace file '{self.trace_file}' is empty."
            )
            resolution = (
                "Enable the trace recorder or provide a dispatch trace to capture "
                "session dependency data required for this check."
            )
            return TestResult(
                summary={
                    "name": "Intra-Session Request Arrival Check",
                    "sections": [
                        {
                            "title": "Trace Required",
                            "results": {
                                "Status": "Skipped",
                                "Reason": reason,
                                "Resolution": resolution,
                            },
                        }
                    ],
                },
                passed=True,
            )

        deviations: List[float] = []
        violations: List[Dict[str, Any]] = []

        session_col = self._get_col("session_id")
        dispatched_col = self._get_col("scheduler_dispatched_at")

        late_threshold = 5.0  # dispatched more than 5 seconds after ready

        for session_id, group in self.merged_df.groupby(session_col):
            req_completion_map = {
                row["request_id"]: row[self._get_col("client_completed_at")]
                for _, row in group.iterrows()
            }

            node_to_request: Dict[Any, int] = {}
            for _, row in group.iterrows():
                ctx = row.get(self._get_col("session_context"), {})
                if ctx and "node_id" in ctx:
                    node_to_request[ctx["node_id"]] = row["request_id"]  # type: ignore

            for _, row in group.iterrows():
                ctx = row.get(self._get_col("session_context"), {})
                if not ctx:
                    continue

                wait_after_ready = ctx["wait_after_ready"]
                parent_nodes = ctx["parent_nodes"]
                session_size = row.get(self._get_col("session_total_requests"))

                parent_finish_times = []
                for node_id in parent_nodes:
                    req_id = node_to_request.get(node_id)
                    if req_id is not None and req_id in req_completion_map:
                        parent_finish_times.append(req_completion_map[req_id])

                if not parent_finish_times:
                    continue

                ready_at = max(parent_finish_times) + wait_after_ready
                actual_dispatch = row[dispatched_col]
                diff = actual_dispatch - ready_at
                deviations.append(diff)

                if diff > late_threshold:
                    violations.append(
                        {
                            "session_id": session_id,
                            "session_size": session_size,
                            "request_id": row["request_id"],
                            "diff": diff,
                            "ready_at": ready_at,
                            "scheduler_dispatched_at": actual_dispatch,
                        }
                    )

        deviations_array = np.array(deviations) if deviations else np.array([])
        passed = len(violations) == 0

        if deviations_array.size > 0:
            stats = {
                "Requests w/ Dependencies": str(len(deviations)),
                "Min": f"{np.min(deviations_array):.4f}s",
                "Mean": f"{np.mean(deviations_array):.4f}s",
                "Median": f"{np.median(deviations_array):.4f}s",
                "P95": f"{np.percentile(deviations_array, 95):.4f}s",
                "P99": f"{np.percentile(deviations_array, 99):.4f}s",
                "Max": f"{np.max(deviations_array):.4f}s",
                "Std Dev": f"{np.std(deviations_array):.4f}s",
            }
        else:
            stats = {"Requests w/ Dependencies": "0"}

        summary = {
            "name": "Intra-Session Request Arrival Check",
            "sections": [
                {
                    "title": "Description",
                    "results": {
                        "Metric": "Scheduler delay (actual_dispatch - ready_time)",
                        "Ready Time": "Parent requests completion + wait_after_ready (if available)",
                    },
                },
                {
                    "title": "Deviation Statistics (seconds)",
                    "results": stats,
                },
                {
                    "title": "Violation Info",
                    "results": {
                        "Late Threshold": f"{late_threshold}s (dispatched >{late_threshold}s after ready)",
                        "Violations": str(len(violations)),
                    },
                },
            ],
        }

        if violations:
            summary["sections"].append(
                {
                    "title": "Violation Details (first 5)",
                    "results": {
                        f"#{i+1}": f"session={v['session_id']} (size={v.get('session_size','?')}), diff={v['diff']:.4f}s"
                        for i, v in enumerate(violations[:5])
                    },
                }
            )

        return TestResult(summary=summary, passed=passed)

    def check_session_concurrency(self) -> TestResult:
        """
        Verify that the number of concurrent sessions tracks the target,
        respecting the ramp-up period.

        A violation occurs when concurrency exceeds the expected limit at any point.
        During ramp-up: limit is linearly interpolated from 0 to target.
        After ramp-up (steady-state): limit is target_concurrent_sessions.
        """
        session_col = self._get_col("session_id")
        dispatched_col = self._get_col("scheduler_dispatched_at")

        # Get benchmark end time (last request completion)
        all_ends = self.merged_df[dispatched_col] + self.merged_df["end_to_end_latency"]
        benchmark_end_time = all_ends.max()

        intervals = []
        for session_id, group in self.merged_df.groupby(session_col):
            starts = group[dispatched_col]
            ends = group[dispatched_col] + group["end_to_end_latency"]

            s_start = starts.min()

            # Check if all requests in session were dispatched
            # session_size tells us expected total requests
            session_size_col = self._get_col("session_size")
            if session_size_col in group.columns:
                expected_requests = group[session_size_col].iloc[0]
                actual_requests = len(group)

                if actual_requests < expected_requests:
                    # Session is still active (waiting for more requests)
                    # Consider it active until benchmark end
                    s_end = benchmark_end_time
                else:
                    # All requests dispatched, use last completion time
                    s_end = ends.max()
            else:
                # No session_size info, fall back to request-based measurement
                s_end = ends.max()

            intervals.append((s_start, s_end))

        # Sweep line to track concurrency over time
        events = []
        for s, e in intervals:
            events.append((s, 1))
            events.append((e, -1))

        events.sort()

        target = self.config.traffic_scheduler.target_concurrent_sessions  # type: ignore
        rampup_seconds = float(self.config.traffic_scheduler.rampup_seconds)  # type: ignore

        current_concurrency = 0
        ramp_violations = []
        steady_violations = []
        time_series = []  # [(time, concurrency)]

        for t, change in events:
            current_concurrency += change
            time_series.append((t, current_concurrency))

            # Compute limit at this time
            if rampup_seconds > 0 and t < rampup_seconds:
                limit = int(target * (t / rampup_seconds))
                phase_violations = ramp_violations
            else:
                limit = target
                phase_violations = steady_violations

            if current_concurrency > limit:
                phase_violations.append(
                    {"time": t, "concurrency": current_concurrency, "limit": limit}
                )

        violations = ramp_violations + steady_violations

        if not time_series:
            return TestResult(
                summary={
                    "name": "Session Concurrency Check",
                    "sections": [
                        {"title": "No Data", "results": {"Status": "No sessions found"}}
                    ],
                },
                passed=True,
            )

        # Compute detailed statistics
        benchmark_start = time_series[0][0]
        benchmark_end = time_series[-1][0]
        total_duration = benchmark_end - benchmark_start

        # Helper to generate smoothed time series
        def smooth_time_series(
            series: List[Tuple[float, int]], target_val: int, tolerance: float
        ) -> List[Tuple[float, int]]:
            """Return a new time series with small dips below target filled."""
            if not series:
                return []

            smoothed = []

            # We iterate intervals. valid interval i is from series[i].t to series[i+1].t
            # series[i].c is the value.
            for i in range(len(series) - 1):
                t_curr, c_curr = series[i]
                t_next, _ = series[i + 1]
                duration = t_next - t_curr

                # Turnaround "gap" usually: c checks below target, duration short.
                # If short duration and below target, bridge it by assuming target.
                # "discounting gaps of a particular size".
                if c_curr < target_val and duration <= tolerance and duration > 0:
                    c_curr = max(c_curr, target_val)

                smoothed.append((t_curr, c_curr))

            # append valid last point
            smoothed.append(series[-1])
            return smoothed

        # Helper to compute time-weighted stats over an interval
        def compute_phase_stats(
            series: List[Tuple[float, int]],
            phase_start: float,
            phase_end: float,
            limit_fn: Callable[[float], float],
        ):
            """Compute time-weighted concurrency stats for a phase."""
            time_below = 0.0
            time_at = 0.0
            time_above = 0.0
            weighted_sum = 0.0
            min_concurrency = float("inf")
            max_concurrency = 0
            duration_by_concurrency: Dict[int, float] = {}

            prev_time = phase_start
            prev_concurrency = 0

            for t, c in series:
                if t <= phase_start:
                    prev_concurrency = c
                    continue
                if t > phase_end:
                    break

                # Account for time from prev_time to min(t, phase_end)
                interval_start = max(prev_time, phase_start)
                interval_end = min(t, phase_end)
                dt = interval_end - interval_start

                if dt > 0:
                    weighted_sum += prev_concurrency * dt
                    min_concurrency = min(min_concurrency, prev_concurrency)
                    max_concurrency = max(max_concurrency, prev_concurrency)
                    duration_by_concurrency[prev_concurrency] = (
                        duration_by_concurrency.get(prev_concurrency, 0.0) + dt
                    )

                    expected_limit = limit_fn((interval_start + interval_end) * 0.5)
                    if prev_concurrency < expected_limit:
                        time_below += dt
                    elif prev_concurrency > expected_limit:
                        time_above += dt
                    else:
                        time_at += dt

                prev_time = t
                prev_concurrency = c

            # Handle final segment to phase_end
            if prev_time < phase_end:
                dt = phase_end - max(prev_time, phase_start)
                if dt > 0:
                    weighted_sum += prev_concurrency * dt
                    min_concurrency = min(min_concurrency, prev_concurrency)
                    max_concurrency = max(max_concurrency, prev_concurrency)
                    duration_by_concurrency[prev_concurrency] = (
                        duration_by_concurrency.get(prev_concurrency, 0.0) + dt
                    )

                    interval_start = max(prev_time, phase_start)
                    interval_end = phase_end
                    expected_limit = limit_fn((interval_start + interval_end) * 0.5)

                    if prev_concurrency < expected_limit:
                        time_below += dt
                    elif prev_concurrency > expected_limit:
                        time_above += dt
                    else:
                        time_at += dt

            phase_duration = phase_end - phase_start
            mean_concurrency = (
                weighted_sum / phase_duration if phase_duration > 0 else 0
            )
            median_concurrency = 0.0
            if phase_duration > 0 and duration_by_concurrency:
                midpoint = phase_duration / 2
                cumulative = 0.0
                for concurrency in sorted(duration_by_concurrency):
                    cumulative += duration_by_concurrency[concurrency]
                    if cumulative >= midpoint:
                        median_concurrency = float(concurrency)
                        break

            return {
                "duration": phase_duration,
                "time_below": time_below,
                "time_at": time_at,
                "time_above": time_above,
                "min": min_concurrency if min_concurrency != float("inf") else 0,
                "max": max_concurrency,
                "mean": mean_concurrency,
                "median": median_concurrency,
            }

        max_observed = max(c for _, c in time_series)
        runtime_timeout = 0.0
        runtime_cfg = getattr(self.config, "runtime", None)
        if runtime_cfg is not None:
            runtime_timeout = float(
                getattr(runtime_cfg, "benchmark_timeout", 0.0) or 0.0
            )
        timeout_cutoff = None
        timeout_triggered = False
        if runtime_timeout > 0:
            timeout_cutoff = benchmark_start + runtime_timeout
            timeout_triggered = benchmark_end >= timeout_cutoff
        post_timeout_context = (
            "After timeout we keep \ndispatching sessions until requests dispatched before \ntimeout finish, but the dispatch "
            "trace stops \nrecording them. Concurrency can look artificially low, so the \n"
            "post-timeout phase isolates this period."
        )

        sections = [
            {
                "title": "Configuration",
                "results": {
                    "Target Concurrency": str(target),
                    "Ramp-Up Period": f"{rampup_seconds:.1f}s",
                    "Benchmark Timeout": (
                        f"{runtime_timeout:.1f}s" if runtime_timeout > 0 else "Disabled"
                    ),
                },
            }
        ]

        def steady_limit_fn(_: float) -> float:
            return target

        def rampup_limit_fn(t: float) -> float:
            if rampup_seconds <= 0:
                return target

            relative_t = t - benchmark_start
            clamped = min(max(relative_t, 0.0), rampup_seconds)
            return int(target * (clamped / rampup_seconds))

        # Ramp-up phase analysis
        if rampup_seconds > 0:
            rampup_end = min(benchmark_start + rampup_seconds, benchmark_end)
            rampup_stats = compute_phase_stats(
                time_series, benchmark_start, rampup_end, rampup_limit_fn
            )
            rampup_duration = rampup_stats["duration"]

            if rampup_duration > 0:
                pct_below = 100.0 * rampup_stats["time_below"] / rampup_duration
                pct_at_or_above = (
                    100.0
                    * (rampup_stats["time_at"] + rampup_stats["time_above"])
                    / rampup_duration
                )

                sections.append(
                    {
                        "title": f"Ramp-Up Phase (0s to {rampup_seconds:.1f}s)",
                        "results": {
                            "Actual Duration": f"{rampup_duration:.2f}s",
                            "Time Below Expected": f"{rampup_stats['time_below']:.2f}s ({pct_below:.1f}%)",
                            "Time At/Above Expected": f"{rampup_stats['time_at'] + rampup_stats['time_above']:.2f}s ({pct_at_or_above:.1f}%)",
                            "Max Observed": str(rampup_stats["max"]),
                        },
                    }
                )

        # Steady-state phase analysis
        steady_start = (
            benchmark_start + rampup_seconds if rampup_seconds > 0 else benchmark_start
        )
        if steady_start < benchmark_end:
            stable_end = benchmark_end
            wind_down_reason = None

            if timeout_triggered and timeout_cutoff is not None:
                stable_end = min(stable_end, timeout_cutoff)
                wind_down_reason = f"Benchmark timeout reached ({runtime_timeout:.2f}s)"
            else:
                threshold = int(target * 0.9)
                # heuristic only used when the timeout-based cutoff is not applicable.
                for t, c in reversed(time_series):
                    if t <= steady_start:
                        break
                    if c >= threshold:
                        stable_end = t
                        break

            if timeout_triggered:
                sections.append(
                    {
                        "title": "Post-Timeout Context",
                        "results": {"Explanation": post_timeout_context},
                    }
                )

            # Compute stats for full steady-state phase (including wind-down)
            full_stats = compute_phase_stats(
                time_series, steady_start, benchmark_end, steady_limit_fn
            )
            full_duration = full_stats["duration"]

            # Compute stats for stable period only (excluding wind-down)
            stable_duration = stable_end - steady_start
            stable_stats = None
            wind_down_duration = max(benchmark_end - stable_end, 0.0)
            if stable_duration > 0 and stable_end < benchmark_end:
                stable_stats = compute_phase_stats(
                    time_series, steady_start, stable_end, steady_limit_fn
                )
            else:
                wind_down_duration = 0.0

            if full_duration > 0:
                pct_below = 100.0 * full_stats["time_below"] / full_duration
                pct_at = 100.0 * full_stats["time_at"] / full_duration
                pct_above = 100.0 * full_stats["time_above"] / full_duration

                sections.append(
                    {
                        "title": f"Steady-State Phase - Full ({rampup_seconds:.1f}s to end)",
                        "results": {
                            "Duration": f"{full_duration:.2f}s",
                            "Time Below Target": f"{full_stats['time_below']:.2f}s ({pct_below:.1f}%)",
                            "Time At Target": f"{full_stats['time_at']:.2f}s ({pct_at:.1f}%)",
                            "Time Above Target": f"{full_stats['time_above']:.2f}s ({pct_above:.1f}%)",
                            "Min Observed": str(full_stats["min"]),
                            "Mean Observed": f"{full_stats['mean']:.2f}",
                            "Median Observed": f"{full_stats['median']:.2f}",
                            "Max Observed": str(full_stats["max"]),
                        },
                    }
                )

            if stable_stats and stable_stats["duration"] > 0:
                s_duration = stable_stats["duration"]
                s_pct_below = 100.0 * stable_stats["time_below"] / s_duration
                s_pct_at = 100.0 * stable_stats["time_at"] / s_duration
                s_pct_above = 100.0 * stable_stats["time_above"] / s_duration

                stable_results = {
                    "Duration": f"{s_duration:.2f}s",
                    "Time Below Target": f"{stable_stats['time_below']:.2f}s ({s_pct_below:.1f}%)",
                    "Time At Target": f"{stable_stats['time_at']:.2f}s ({s_pct_at:.1f}%)",
                    "Time Above Target": f"{stable_stats['time_above']:.2f}s ({s_pct_above:.1f}%)",
                    "Min Observed": str(stable_stats["min"]),
                    "Mean Observed": f"{stable_stats['mean']:.2f}",
                    "Median Observed": f"{stable_stats['median']:.2f}",
                    "Max Observed": str(stable_stats["max"]),
                }
                if wind_down_reason and timeout_triggered:
                    stable_results["Post-Timeout Trigger"] = wind_down_reason

                stable_title_suffix = (
                    f"excludes {wind_down_duration:.1f}s post-timeout"
                    if timeout_triggered
                    else f"excludes {wind_down_duration:.1f}s tail"
                )
                sections.append(
                    {
                        "title": f"Steady-State Phase - Stable ({stable_title_suffix})",
                        "results": stable_results,
                    }
                )

                # Smoothed stable stats
                gap_tolerance = 0.5
                smoothed_series = smooth_time_series(time_series, target, gap_tolerance)
                smoothed_stats = compute_phase_stats(
                    smoothed_series, steady_start, stable_end, steady_limit_fn
                )

                sm_duration = smoothed_stats["duration"]
                if sm_duration > 0:
                    sm_pct_below = 100.0 * smoothed_stats["time_below"] / sm_duration
                    sm_pct_at = 100.0 * smoothed_stats["time_at"] / sm_duration
                    sm_pct_above = 100.0 * smoothed_stats["time_above"] / sm_duration

                    smoothed_title = (
                        f"Steady-State Phase - Stable Smoothed (pre-timeout, ignoring gaps <= {gap_tolerance}s)"
                        if timeout_triggered
                        else f"Steady-State Phase - Stable Smoothed (ignoring gaps <= {gap_tolerance}s)"
                    )
                    smoothed_results = {
                        "Duration": f"{sm_duration:.2f}s",
                        "Time Below Target": f"{smoothed_stats['time_below']:.2f}s ({sm_pct_below:.1f}%)",
                        "Time At Target": f"{smoothed_stats['time_at']:.2f}s ({sm_pct_at:.1f}%)",
                        "Time Above Target": f"{smoothed_stats['time_above']:.2f}s ({sm_pct_above:.1f}%)",
                        "Min Observed": str(smoothed_stats["min"]),
                        "Mean Observed": f"{smoothed_stats['mean']:.2f}",
                        "Median Observed": f"{smoothed_stats['median']:.2f}",
                        "Max Observed": str(smoothed_stats["max"]),
                    }
                    if timeout_triggered:
                        smoothed_results["Post-Timeout Coverage"] = (
                            f"Excluded (timeout at {runtime_timeout:.2f}s)"
                        )
                        if wind_down_reason:
                            smoothed_results["Post-Timeout Trigger"] = wind_down_reason

                    sections.append(
                        {
                            "title": smoothed_title,
                            "results": smoothed_results,
                        }
                    )

                if timeout_triggered and wind_down_duration > 0:
                    wind_down_stats = compute_phase_stats(
                        time_series, stable_end, benchmark_end, steady_limit_fn
                    )
                    wd_duration = wind_down_stats["duration"]
                    if wd_duration > 0:
                        wd_pct_below = (
                            100.0 * wind_down_stats["time_below"] / wd_duration
                        )
                        wd_pct_at = 100.0 * wind_down_stats["time_at"] / wd_duration
                        wd_pct_above = (
                            100.0 * wind_down_stats["time_above"] / wd_duration
                        )

                        wind_down_results = {
                            "Duration": f"{wd_duration:.2f}s",
                            "Time Below Target": f"{wind_down_stats['time_below']:.2f}s ({wd_pct_below:.1f}%)",
                            "Time At Target": f"{wind_down_stats['time_at']:.2f}s ({wd_pct_at:.1f}%)",
                            "Time Above Target": f"{wind_down_stats['time_above']:.2f}s ({wd_pct_above:.1f}%)",
                            "Min Observed": str(wind_down_stats["min"]),
                            "Mean Observed": f"{wind_down_stats['mean']:.2f}",
                            "Median Observed": f"{wind_down_stats['median']:.2f}",
                            "Max Observed": str(wind_down_stats["max"]),
                            "Explanation": post_timeout_context,
                        }
                        if wind_down_reason:
                            wind_down_results["Trigger"] = wind_down_reason

                        sections.append(
                            {
                                "title": "Post-Timeout Phase",
                                "results": wind_down_results,
                            }
                        )

        # Overall statistics
        sections.append(
            {
                "title": "Overall Statistics",
                "results": {
                    "Total Duration": f"{total_duration:.2f}s",
                    "Max Observed Concurrency": str(max_observed),
                },
            }
        )

        # Violation info
        sections.append(
            {
                "title": "Violation Info",
                "results": {
                    "Definition": "Concurrency exceeds phase-specific limit at any point",
                    "Ramp-Up Limit": f"Linear from 0 to {target}",
                    "Steady-State Limit": str(target),
                    "Ramp-Up Violations": str(len(ramp_violations)),
                    "Steady-State Violations": str(len(steady_violations)),
                    "Total Violations": str(len(violations)),
                },
            }
        )

        def append_violation_sections(
            title: str, violation_list: List[Dict[str, float]]
        ):
            if not violation_list:
                return

            overshoots = np.array(
                [v["concurrency"] - v["limit"] for v in violation_list]
            )
            sections.append(
                {
                    "title": f"{title} Violation Distribution (overshoot = concurrency - limit)",
                    "results": {
                        "Count": str(len(violation_list)),
                        "Min Overshoot": str(int(np.min(overshoots))),
                        "Mean Overshoot": f"{np.mean(overshoots):.2f}",
                        "Median Overshoot": f"{np.median(overshoots):.1f}",
                        "P95 Overshoot": f"{np.percentile(overshoots, 95):.1f}",
                        "Max Overshoot": str(int(np.max(overshoots))),
                    },
                }
            )

            violation_times = np.array([v["time"] for v in violation_list])
            sections.append(
                {
                    "title": f"{title} Violation Time Distribution (seconds)",
                    "results": {
                        "First Violation": f"{np.min(violation_times):.2f}s",
                        "Last Violation": f"{np.max(violation_times):.2f}s",
                        "Median Time": f"{np.median(violation_times):.2f}s",
                    },
                }
            )

            sections.append(
                {
                    "title": f"{title} Violation Details (first 5)",
                    "results": {
                        f"#{i+1}": f"t={v['time']:.2f}s, concurrency={v['concurrency']}, limit={v['limit']}"
                        for i, v in enumerate(violation_list[:5])
                    },
                }
            )

        append_violation_sections("Ramp-Up", ramp_violations)
        append_violation_sections("Steady-State", steady_violations)

        passed = len(violations) == 0

        summary = {"name": "Session Concurrency Check", "sections": sections}
        return TestResult(summary=summary, passed=passed)

    def check_session_dispatch_rate(self) -> TestResult:
        """
        Verify that the actual session dispatch rate matches the expected arrival rate.

        Calculated as: (N - 1) / (last_start_time - first_start_time)
        Matches against the configured arrival_rate (for Poisson/Gamma) or 1/interval (for Fixed).
        """
        # Determine expected rate
        scheduler_config = self.config.traffic_scheduler
        interval_config = scheduler_config.interval_generator  # type: ignore
        generator_type = interval_config.get_type()

        expected_rate = 0.0
        if generator_type in [
            IntervalGeneratorType.POISSON,
            IntervalGeneratorType.GAMMA,
        ]:
            expected_rate = float(interval_config.arrival_rate)
        elif generator_type == IntervalGeneratorType.FIXED:
            interval = float(interval_config.interval)
            expected_rate = 1.0 / interval if interval > 0 else 0.0

        if expected_rate <= 0:
            return TestResult(
                summary={
                    "name": "Session Dispatch Rate Check",
                    "sections": [
                        {
                            "title": "Invalid Configuration",
                            "results": {
                                "Status": "Skipped",
                                "Reason": f"Expected rate <= 0 (type={generator_type.name})",
                            },
                        }
                    ],
                },
                passed=True,
            )

        # Get session start times
        session_col = self._get_col("session_id")
        dispatched_col = self._get_col("scheduler_dispatched_at")

        if self.merged_df.empty:
            return TestResult(
                summary={
                    "name": "Session Dispatch Rate Check",
                    "sections": [
                        {"title": "No Data", "results": {"Status": "No sessions found"}}
                    ],
                },
                passed=True,
            )

        # first request dispatch time for each session (min for multi root sessions)
        session_starts = self.merged_df.groupby(session_col)[dispatched_col].min()
        start_times = np.sort(np.array(session_starts.values))

        if len(start_times) < 2:
            return TestResult(
                summary={
                    "name": "Session Dispatch Rate Check",
                    "sections": [
                        {
                            "title": "Insufficient Data",
                            "results": {
                                "Status": "Skipped",
                                "Reason": f"Only {len(start_times)} session(s) found, need at least 2",
                            },
                        }
                    ],
                },
                passed=True,
            )

        duration = start_times[-1] - start_times[0]
        count = len(start_times) - 1
        actual_rate = count / duration if duration > 0 else 0.0

        inter_arrival_times = np.diff(start_times)
        if len(inter_arrival_times) == 0:
            pass

        error_pct = (
            abs(actual_rate - expected_rate) / expected_rate * 100.0
            if expected_rate > 0
            else 0.0
        )
        threshold_pct = 15.0

        passed = error_pct <= threshold_pct

        stats = {
            "Total Sessions": str(len(start_times)),
            "Measurement Duration": f"{duration:.4f}s",
            "Expected Rate": f"{expected_rate:.4f} sessions/sec",
            "Actual Rate": f"{actual_rate:.4f} sessions/sec",
            "Error": f"{error_pct:.2f}%",
            "Threshold": f"{threshold_pct:.1f}%",
        }

        inter_arrival_stats = {
            "Min": f"{np.min(inter_arrival_times):.4f}s",
            "Mean": f"{np.mean(inter_arrival_times):.4f}s",
            "Median": f"{np.median(inter_arrival_times):.4f}s",
            "P95": f"{np.percentile(inter_arrival_times, 95):.4f}s",
            "P99": f"{np.percentile(inter_arrival_times, 99):.4f}s",
            "Max": f"{np.max(inter_arrival_times):.4f}s",
            "Std Dev": f"{np.std(inter_arrival_times):.4f}s",
        }

        summary = {
            "name": "Session Dispatch Rate Check",
            "sections": [
                {
                    "title": "Rate Statistics",
                    "results": stats,
                },
                {
                    "title": "Inter-Arrival Time Statistics",
                    "results": inter_arrival_stats,
                },
            ],
        }

        if not passed:
            summary["sections"].append(
                {
                    "title": "Failure Reason",
                    "results": {
                        "Result": "FAILED",
                        "Explanation": f"Actual rate deviates from expected by > {threshold_pct}%",
                    },
                }
            )

        return TestResult(summary=summary, passed=passed)

    def check_prompt_length(self) -> TestResult:
        """
        Verify that the generated prompt length matches the target prompt length.
        Checks deviation: num_delta_prompt_tokens - target_num_delta_prompt_tokens.
        """
        target_col = self._get_col("target_num_delta_prompt_tokens")
        actual_col_delta = self._get_col("num_delta_prompt_tokens")

        # Fallback to checking num_prompt_tokens if delta not available (legacy)
        # But user specifically asked for delta. The generators populate "target_prompt_tokens".
        # Evaluator maps this to "target_num_delta_prompt_tokens".
        # "num_delta_prompt_tokens" is from channel metrics.

        if (
            target_col not in self.merged_df.columns
            or actual_col_delta not in self.merged_df.columns
        ):
            return TestResult(
                summary={
                    "name": "Prompt Length Check",
                    "sections": [
                        {
                            "title": "Missing Data",
                            "results": {
                                "Status": "Skipped",
                                "Reason": f"Columns '{target_col}' or '{actual_col_delta}' missing.",
                            },
                        }
                    ],
                },
                passed=True,
            )

        # Filter rows where target is present (non-null and > 0)
        df = self.merged_df.dropna(subset=[target_col, actual_col_delta])
        if df.empty:
            return TestResult(
                summary={
                    "name": "Prompt Length Check",
                    "sections": [
                        {
                            "title": "No Target Data",
                            "results": {
                                "Status": "Skipped",
                                "Reason": "No requests found with target prompt tokens specified.",
                            },
                        }
                    ],
                },
                passed=True,
            )

        df["prompt_len_diff"] = df[actual_col_delta] - df[target_col]
        deviations = df["prompt_len_diff"].to_numpy()

        # Relaxed check with threshold
        threshold = 15.0
        violations = df[df["prompt_len_diff"].abs() > threshold]
        violation_count = len(violations)
        total_count = len(df)

        passed = violation_count == 0

        stats = {
            "Total Requests Checked": str(total_count),
            "Exact Matches": f"{len(df[df['prompt_len_diff'] == 0])} ({100.0 * len(df[df['prompt_len_diff'] == 0]) / total_count:.1f}%)",
            "Mismatches (All)": f"{len(df[df['prompt_len_diff'] != 0])} ({100.0 * len(df[df['prompt_len_diff'] != 0]) / total_count:.1f}%)",
            "Violations (> +/-15)": f"{violation_count} ({100.0 * violation_count / total_count:.1f}%)",
        }

        if deviations.size > 0:
            stats.update(
                {
                    "Min Deviation": f"{np.min(deviations):.1f}",
                    "Mean Deviation": f"{np.mean(deviations):.2f}",
                    "Median Deviation": f"{np.median(deviations):.1f}",
                    "P95 Deviation": f"{np.percentile(deviations, 95):.1f}",
                    "P99 Deviation": f"{np.percentile(deviations, 99):.1f}",
                    "Max Deviation": f"{np.max(deviations):.1f}",
                    "Std Dev": f"{np.std(deviations):.2f}",
                }
            )

        summary = {
            "name": "Prompt Length Check",
            "sections": [
                {
                    "title": "Description",
                    "results": {
                        "Metric": "Prompt Length Deviation (Actual - Target)",
                        "Target": "Specified target_prompt_tokens",
                        "Threshold": f"<= +/- {threshold}",
                    },
                },
                {
                    "title": "Statistics",
                    "results": stats,
                },
            ],
        }

        if violation_count > 0:
            # Add top violations (largest absolute difference)
            violations_df = cast(pd.DataFrame, violations.copy())
            violations_df["abs_diff"] = np.abs(violations_df["prompt_len_diff"])
            top_violations = violations_df.nlargest(5, "abs_diff")

            summary["sections"].append(
                {
                    "title": "Top Violations (Largest Absolute Difference)",
                    "results": {
                        f"#{i+1}": f"req={row['request_id']}, target={row[target_col]}, actual={row[actual_col_delta]}, diff={row['prompt_len_diff']}"
                        for i, (_, row) in enumerate(top_violations.iloc[:5].iterrows())
                    },
                }
            )

        return TestResult(summary=summary, passed=passed)

    def check_output_length(self) -> TestResult:
        """
        Verify that the generated output length matches the requested output length.
        Checks deviation: num_output_tokens - num_requested_output_tokens.
        """
        target_col = self._get_col("num_requested_output_tokens")
        actual_col = self._get_col("num_output_tokens")

        if (
            target_col not in self.merged_df.columns
            or actual_col not in self.merged_df.columns
        ):
            return TestResult(
                summary={
                    "name": "Output Length Check",
                    "sections": [
                        {
                            "title": "Missing Data",
                            "results": {
                                "Status": "Skipped",
                                "Reason": f"Columns '{target_col}' or '{actual_col}' missing.",
                            },
                        }
                    ],
                },
                passed=True,
            )

        # Filter rows where target is present (non-null and > 0)
        df = self.merged_df.dropna(subset=[target_col, actual_col])
        if df.empty:
            return TestResult(
                summary={
                    "name": "Output Length Check",
                    "sections": [
                        {
                            "title": "No Data",
                            "results": {
                                "Status": "Skipped",
                                "Reason": "No requests found with requested output tokens specified.",
                            },
                        }
                    ],
                },
                passed=True,
            )

        df["output_len_diff"] = df[actual_col] - df[target_col]
        deviations = df["output_len_diff"].to_numpy()

        # Relaxed check with threshold
        threshold = 15.0  # Allow some deviation
        violations = df[df["output_len_diff"].abs() > threshold]
        violation_count = len(violations)
        total_count = len(df)

        passed = violation_count == 0

        stats = {
            "Total Requests Checked": str(total_count),
            "Exact Matches": f"{len(df[df['output_len_diff'] == 0])} ({100.0 * len(df[df['output_len_diff'] == 0]) / total_count:.1f}%)",
            "Mismatches (All)": f"{len(df[df['output_len_diff'] != 0])} ({100.0 * len(df[df['output_len_diff'] != 0]) / total_count:.1f}%)",
            "Violations (> +/-15)": f"{violation_count} ({100.0 * violation_count / total_count:.1f}%)",
        }

        if deviations.size > 0:
            stats.update(
                {
                    "Min Deviation": f"{np.min(deviations):.1f}",
                    "Mean Deviation": f"{np.mean(deviations):.2f}",
                    "Median Deviation": f"{np.median(deviations):.1f}",
                    "P95 Deviation": f"{np.percentile(deviations, 95):.1f}",
                    "P99 Deviation": f"{np.percentile(deviations, 99):.1f}",
                    "Max Deviation": f"{np.max(deviations):.1f}",
                    "Std Dev": f"{np.std(deviations):.2f}",
                }
            )

        summary = {
            "name": "Output Length Check",
            "sections": [
                {
                    "title": "Description",
                    "results": {
                        "Metric": "Output Length Deviation (Actual - Requested)",
                        "Target": "num_requested_output_tokens",
                        "Threshold": f"<= +/- {threshold}",
                    },
                },
                {
                    "title": "Statistics",
                    "results": stats,
                },
            ],
        }

        if violation_count > 0:
            # Add top violations (largest absolute difference)
            violations_df = cast(pd.DataFrame, violations.copy())
            violations_df["abs_diff"] = np.abs(violations_df["output_len_diff"])
            top_violations = violations_df.nlargest(5, "abs_diff")

            summary["sections"].append(
                {
                    "title": "Top Violations (Largest Absolute Difference)",
                    "results": {
                        f"#{i+1}": f"req={row['request_id']}, target={row[target_col]}, actual={row[actual_col]}, diff={row['output_len_diff']}"
                        for i, (_, row) in enumerate(top_violations.iloc[:5].iterrows())
                    },
                }
            )

        return TestResult(summary=summary, passed=passed)

    def check_lifecycle_timing_delays(self) -> TestResult:
        """
        Report statistics on lifecycle timing delays between pipeline stages.

        Computes percentiles for:
        - scheduler_ready_at -> scheduler_dispatched_at (dispatch delay)
        - scheduler_dispatched_at -> client_picked_up_at (queue wait time)
        - client_completed_at -> result_processed_at (result processing delay)
        """

        def compute_delay_stats(delays: np.ndarray, name: str) -> Dict[str, str]:
            """Compute percentile statistics for a delay array."""
            if delays.size == 0:
                return {f"{name} (no data)": "N/A"}

            # Filter out NaN values
            valid_delays = delays[~np.isnan(delays)]
            if valid_delays.size == 0:
                return {f"{name} (no valid data)": "N/A"}

            return {
                "Count": str(valid_delays.size),
                "Min": f"{np.min(valid_delays):.4f}s",
                "Mean": f"{np.mean(valid_delays):.4f}s",
                "Median": f"{np.median(valid_delays):.4f}s",
                "P95": f"{np.percentile(valid_delays, 95):.4f}s",
                "P99": f"{np.percentile(valid_delays, 99):.4f}s",
                "Max": f"{np.max(valid_delays):.4f}s",
                "Std Dev": f"{np.std(valid_delays):.4f}s",
            }

        sections = []

        # Check if lifecycle columns exist
        ready_col = "scheduler_ready_at"
        dispatched_col = "scheduler_dispatched_at"
        pickup_col = "client_picked_up_at"
        completed_col = "client_completed_at"
        processed_col = "result_processed_at"

        missing_cols = []
        for col in [
            ready_col,
            dispatched_col,
            pickup_col,
            completed_col,
            processed_col,
        ]:
            if col not in self.metrics_df.columns:
                missing_cols.append(col)

        if missing_cols:
            return TestResult(
                summary={
                    "name": "Lifecycle Timing Delays Check",
                    "sections": [
                        {
                            "title": "Missing Data",
                            "results": {
                                "Missing Columns": ", ".join(missing_cols),
                                "Resolution": "Re-run benchmark with updated code that captures lifecycle timestamps.",
                            },
                        }
                    ],
                },
                passed=True,  # Not a failure, just missing data
            )

        df = self.metrics_df.copy()

        # 1. Ready-to-Dispatch delay
        ready_to_dispatch = (df[dispatched_col] - df[ready_col]).values
        sections.append(
            {
                "title": "Ready-to-Dispatch Delay (scheduler_dispatched_at - scheduler_ready_at)",
                "results": compute_delay_stats(ready_to_dispatch, "Delay"),
            }
        )

        # 2. Dispatch-to-Pickup delay (queue wait time)
        dispatch_to_pickup = (df[pickup_col] - df[dispatched_col]).values
        sections.append(
            {
                "title": "Dispatch-to-Pickup Delay (client_picked_up_at - scheduler_dispatched_at)",
                "results": compute_delay_stats(dispatch_to_pickup, "Delay"),
            }
        )

        # 3. Completion-to-Result-Processing delay
        completion_to_processing = (df[processed_col] - df[completed_col]).values
        sections.append(
            {
                "title": "Completion-to-Result-Processing Delay (result_processed_at - client_completed_at)",
                "results": compute_delay_stats(completion_to_processing, "Delay"),
            }
        )

        return TestResult(
            summary={
                "name": "Lifecycle Timing Delays Check",
                "sections": sections,
            },
            passed=True,  # No pass/fail criteria for now
        )



from dataclasses import field

from veeksha.config.core.base_poly_config import BasePolyConfig
from veeksha.config.core.frozen_dataclass import frozen_dataclass
from veeksha.config.generator.length import (
    BaseLengthGeneratorConfig,
    UniformLengthGeneratorConfig,
)
from veeksha.types import ChannelModality, LengthGeneratorType
from veeksha.types.base_registry import BaseRegistry


@frozen_dataclass
class BaseChannelGeneratorConfig(BasePolyConfig):
    pass


@frozen_dataclass
class TextChannelGeneratorConfig(BaseChannelGeneratorConfig):

    body_length_generator: BaseLengthGeneratorConfig = field(
        default_factory=UniformLengthGeneratorConfig,
        metadata={
            "help": f"The generator for the body length. {LengthGeneratorType.help_str()}"
        },
    )
    output_length_generator: BaseLengthGeneratorConfig = field(
        default_factory=UniformLengthGeneratorConfig,
        metadata={
            "help": f"The generator for the output length. {LengthGeneratorType.help_str()}"
        },
    )
    shared_prefix_ratio: float = field(
        default=0.0,
        metadata={
            "help": "Fraction of prompt tokens to use as shared prefix for root requests (0.0-1.0)"
        },
    )
    shared_prefix_probability: float = field(
        default=1.0,
        metadata={
            "help": "Probability that a root request uses shared prefix (0.0-1.0)"
        },
    )

    @classmethod
    def get_type(cls):
        return ChannelModality.TEXT


@frozen_dataclass
class ImageChannelGeneratorConfig(BaseChannelGeneratorConfig):

    def __post_init__(self):
        raise NotImplementedError("ImageChannelConfig is not implemented")

    @classmethod
    def get_type(cls):
        return ChannelModality.IMAGE


@frozen_dataclass
class AudioChannelGeneratorConfig(BaseChannelGeneratorConfig):
    def __post_init__(self):
        raise NotImplementedError("AudioChannelConfig is not implemented")

    @classmethod
    def get_type(cls):
        return ChannelModality.AUDIO


@frozen_dataclass
class VideoChannelGeneratorConfig(BaseChannelGeneratorConfig):

    def __post_init__(self):
        raise NotImplementedError("VideoChannelConfig is not implemented")

    @classmethod
    def get_type(cls):
        return ChannelModality.VIDEO


# channel registry
class ChannelGeneratorRegistry(BaseRegistry):
    @classmethod
    def get_key_from_str(cls, key_str: str) -> ChannelModality:
        return ChannelModality.from_str(key_str)  # type: ignore


ChannelGeneratorRegistry.register(ChannelModality.TEXT, TextChannelGeneratorConfig)
ChannelGeneratorRegistry.register(ChannelModality.IMAGE, ImageChannelGeneratorConfig)
ChannelGeneratorRegistry.register(ChannelModality.AUDIO, AudioChannelGeneratorConfig)
ChannelGeneratorRegistry.register(ChannelModality.VIDEO, VideoChannelGeneratorConfig)



from enum import IntEnum


class BaseIntEnum(IntEnum):
    def __str__(self):
        return self.name.lower()

    @classmethod
    def from_str(cls, string):
        return cls[string.upper()]

    @classmethod
    def help_str(cls):
        return f"Available: {', '.join([m.name.lower() for m in cls])}."



from dataclasses import field
from typing import Optional

from veeksha.config.core.frozen_dataclass import frozen_dataclass


@frozen_dataclass(allow_from_file=True)
class WandbConfig:
    """Configuration for Weights & Biases logging.

    Attributes:
        enabled: If True, the benchmark runner will initialize a wandb run and
            upload selected outputs/metrics.
        project: WandB project name. If unset, wandb will fall back to its
            standard resolution (e.g., `WANDB_PROJECT`).
        entity: WandB entity (user/team). Optional.
        group: WandB group name for logically-related runs (e.g., capacity search
            attempts or CLI sweeps). Optional.
        run_name: Optional run name. If unset, veeksha will default to the
            resolved output directory name for uniqueness.
        tags: Optional list of tags to attach to the run.
        notes: Optional human-readable notes.
        mode: Optional wandb mode override ("online", "offline", "disabled").
            If unset, wandb uses `WANDB_MODE` and its defaults.
        log_artifacts: If True, upload selected output files as a wandb Artifact.
    """

    enabled: bool = field(
        default=False, metadata={"help": "Enable Weights & Biases logging."}
    )
    project: Optional[str] = field(
        default=None,
        metadata={"help": "WandB project name (or set WANDB_PROJECT)."},
    )
    entity: Optional[str] = field(
        default=None,
        metadata={"help": "WandB entity (team/user). Optional."},
    )
    group: Optional[str] = field(
        default=None,
        metadata={"help": "WandB group name (for sweeps/capacity-search)."},
    )
    run_name: Optional[str] = field(
        default=None,
        metadata={"help": "WandB run name. Defaults to the resolved output dir name."},
    )
    tags: list[str] = field(
        default_factory=list,
        metadata={"help": "List of WandB tags to attach to the run."},
    )
    notes: Optional[str] = field(
        default=None,
        metadata={"help": "Optional WandB notes for this run."},
    )
    mode: Optional[str] = field(
        default=None,
        metadata={
            "help": "Optional wandb mode override: 'online', 'offline', or 'disabled'."
        },
    )
    log_artifacts: bool = field(
        default=True,
        metadata={"help": "Upload selected output files as a wandb Artifact."},
    )



"""CLI runner for `python -m veeksha.new`.

We keep the module entrypoint (`veeksha/new/__main__.py`) intentionally thin and
centralize orchestration logic here for readability and reuse.
"""

from __future__ import annotations

import os
from collections import defaultdict
from dataclasses import replace
from datetime import datetime, timezone

from veeksha.benchmark import manage_benchmark_run
from veeksha.config.benchmark import BenchmarkConfig
from veeksha.logger import init_logger
from veeksha.sweep_summary import write_sweep_summary
from veeksha.wandb_integration import dedup_tags, maybe_log_sweep_summary

logger = init_logger(__name__)


class BenchmarkCliRunner:
    """Runs one or more `BenchmarkConfig`s produced by the CLI/YAML loader."""

    def __init__(self, benchmark_configs: list[BenchmarkConfig]):
        self._benchmark_configs = benchmark_configs

    @classmethod
    def from_cli(cls) -> "BenchmarkCliRunner":
        return cls(BenchmarkConfig.create_from_cli_args())

    def run_all(self) -> None:
        """Run all benchmark configs, grouping sweeps by base output directory."""
        configs_by_base_dir: dict[str, list[BenchmarkConfig]] = defaultdict(list)
        for cfg in self._benchmark_configs:
            configs_by_base_dir[cfg.output_dir].append(cfg)

        for base_output_dir, configs in configs_by_base_dir.items():
            if len(configs) > 1:
                self._run_sweep(base_output_dir=base_output_dir, configs=configs)
            else:
                self._run_single(configs[0])

    def _run_single(self, cfg: BenchmarkConfig) -> None:
        manage_benchmark_run(cfg)

    def _run_sweep(
        self, *, base_output_dir: str, configs: list[BenchmarkConfig]
    ) -> None:
        sweep_timestamp = datetime.now(timezone.utc).strftime("%d:%m:%Y-%H:%M:%S")
        sweep_dir = os.path.join(base_output_dir, f"sweep_{sweep_timestamp}")
        os.makedirs(sweep_dir, exist_ok=True)
        sweep_group = f"sweep-{os.path.basename(sweep_dir.rstrip('/'))}"

        all_run_dirs: list[str] = []

        for cfg in configs:
            wandb_cfg = cfg.wandb
            if wandb_cfg.enabled:
                wandb_cfg = replace(
                    wandb_cfg,
                    group=wandb_cfg.group or sweep_group,
                    tags=dedup_tags([*wandb_cfg.tags, "sweep"]),
                )

            updated_cfg = replace(cfg, output_dir=sweep_dir, wandb=wandb_cfg)
            manage_benchmark_run(updated_cfg)
            # manage_benchmark_run mutates output_dir to the resolved run dir
            all_run_dirs.append(updated_cfg.output_dir)

        write_sweep_summary(sweep_dir, all_run_dirs)

        first = configs[0]
        if first.wandb.enabled:
            maybe_log_sweep_summary(
                sweep_dir=sweep_dir,
                wandb_cfg=first.wandb,
                group=first.wandb.group or sweep_group,
            )


def main() -> None:
    BenchmarkCliRunner.from_cli().run_all()



from veeksha.config.generator.session import SyntheticSessionGeneratorConfig
from veeksha.core.request import Request
from veeksha.core.seeding import SeedManager
from veeksha.core.session import Session
from veeksha.core.session_graph import get_node_ids, is_root, parents
from veeksha.core.tokenizer import TokenizerProvider
from veeksha.generator.channel.registry import ChannelGeneratorRegistry
from veeksha.generator.session.base import BaseSessionGenerator
from veeksha.generator.session_graph.registry import SessionGraphGeneratorRegistry


class SyntheticSessionGenerator(BaseSessionGenerator):
    def __init__(
        self,
        config: SyntheticSessionGeneratorConfig,
        seed_manager: SeedManager,
        tokenizer_provider: TokenizerProvider,
        append_min_tokens_instruction: bool = False,
    ):
        self.config = config
        self.seed_manager = seed_manager
        self.tokenizer_provider = tokenizer_provider
        self.append_min_tokens_instruction = append_min_tokens_instruction

        # get generators
        self.channels = {}
        for channel in self.config.channels:
            tokenizer_handle = self.tokenizer_provider.for_modality(channel.get_type())
            channel_kwargs = {
                "seed_manager": self.seed_manager.child(
                    f"channel_{channel.get_type()}"
                ),
                "tokenizer_handle": tokenizer_handle,
            }
            if channel.get_type() == ChannelGeneratorRegistry.get_key_from_str("text"):
                channel_kwargs["append_min_tokens_instruction"] = (
                    self.append_min_tokens_instruction
                )

            self.channels[channel.get_type()] = ChannelGeneratorRegistry.get(
                channel.get_type(),
                channel,
                **channel_kwargs,
            )
        self.session_graph_generator = SessionGraphGeneratorRegistry.get(
            self.config.session_graph.get_type(),
            self.config.session_graph,
            seed_manager=seed_manager.child("session_graph"),
        )

        self.current_session_id = 0  # incremental global session id
        self.current_request_id = 0  # incremental global request id

    def generate_session(self) -> Session:
        session_graph = self.session_graph_generator.generate_session_graph()
        requests = {}

        for node_id in get_node_ids(session_graph):
            # get content
            channels = {}
            for channel_type, channel in self.channels.items():
                channels[channel_type] = channel.generate_content(
                    is_root=is_root(session_graph, node_id)
                )

            incoming_edges = parents(session_graph, node_id)
            history_parents = [e for e in incoming_edges if e.is_history_parent]
            session_context = {
                "node_id": node_id,
                "wait_after_ready": session_graph.nodes[node_id].wait_after_ready,
                "parent_nodes": [e.src for e in incoming_edges],
                "history_parent": history_parents[0].src if history_parents else None,
            }

            request = Request(
                id=self.current_request_id,
                channels=channels,
                session_context=session_context,
            )
            requests[node_id] = request
            self.current_request_id += 1
        session = Session(
            id=self.current_session_id,
            session_graph=session_graph,
            requests=requests,
        )
        self.current_session_id += 1
        return session

    def capacity(self) -> int:
        return -1



"""E2E Test for Managed Server Benchmark."""

import json
import threading
import time
from http.server import BaseHTTPRequestHandler, HTTPServer
from unittest.mock import MagicMock, patch

import pytest
import yaml

from veeksha.benchmark import manage_benchmark_run
from veeksha.config.benchmark import BenchmarkConfig
from veeksha.config.utils import create_class_from_dict
from veeksha.core.request_content import TextChannelRequestContent

SAMPLE_CONFIG_PATH = "veeksha/sample_configs/managed_server.yml"

class MockOpenAIHandler(BaseHTTPRequestHandler):
    def do_POST(self):
        content_len = int(self.headers.get('Content-Length', 0))
        body = self.rfile.read(content_len)
        data = json.loads(body)
        
        # Determine number of tokens requested
        max_tokens = data.get("max_completion_tokens", 10)
        
        response = {
            "id": "chatcmpl-mock",
            "object": "chat.completion",
            "created": int(time.time()),
            "choices": [{
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": "mock " * max_tokens
                },
                "finish_reason": "stop"
            }],
            "usage": {
                "prompt_tokens": 10,
                "completion_tokens": max_tokens,
                "total_tokens": 10 + max_tokens
            }
        }
        
        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(json.dumps(response).encode('utf-8'))
    
    def log_message(self, format, *args):
        pass # Silence logs

@pytest.fixture
def mock_openai_server():
    server = HTTPServer(('localhost', 0), MockOpenAIHandler)
    thread = threading.Thread(target=server.serve_forever)
    thread.daemon = True
    thread.start()
    yield server
    server.shutdown()

# @pytest.mark.e2e
def test_managed_server_benchmark(mock_openai_server, tmp_path) -> None:
    # 1. Load config
    with open(SAMPLE_CONFIG_PATH, "r") as f:
        config_dict = yaml.safe_load(f)
    
    # 2. Modify config for test
    config_dict["runtime"]["max_sessions"] = 5 
    config_dict["runtime"]["benchmark_timeout"] = 10
    config_dict["output_dir"] = str(tmp_path)
    
    # 3. Create BenchmarkConfig
    benchmark_config = create_class_from_dict(BenchmarkConfig, config_dict)
    
    # 4. Mock managed_server context manager
    port = mock_openai_server.server_port
    server_info = {
        "api_base": f"http://localhost:{port}/v1",
        "api_key": "dummy",
        "model": config_dict["server"]["model"]
    }
    
    mock_ctx = MagicMock()
    mock_ctx.__enter__.return_value = server_info
    mock_ctx.__exit__.return_value = None
    
    # We patch the call to managed_server in veeksha.benchmark
    with patch("veeksha.benchmark.managed_server", return_value=mock_ctx) as mocked_ms, \
         patch("veeksha.benchmark.build_hf_tokenizer_handle_from_model") as mock_build_tok, \
         patch("veeksha.core.tokenizer.build_hf_tokenizer_handle_from_model") as mock_build_tok_core:
        
        # Mock tokenizer handle
        mock_handle = MagicMock()
        # Mock encode to return a list of token IDs
        mock_handle.encode.return_value = [1] * 10
        # Mock decode
        mock_handle.decode.return_value = "mock_text"
        mock_build_tok.return_value = mock_handle
        mock_build_tok_core.return_value = mock_handle

        # Stub the channel generator to avoid heavy prompt generation
        mock_channel_gen = MagicMock()
        mock_channel_gen.generate_content.return_value = TextChannelRequestContent(
            input_text="mock prompt",
            target_output_tokens=1,
            target_prompt_tokens=5,
        )

        with patch(
            "veeksha.generator.channel.registry.ChannelGeneratorRegistry.get",
            return_value=mock_channel_gen,
        ):

            # 5. Run
            result = manage_benchmark_run(benchmark_config)

            # 6. Verify
            mocked_ms.assert_called_once()
            assert result.metrics.get("Successful Sessions", 0) == 5
            assert result.metrics.get("Number of Completed Requests", 0) > 0
            assert result.metrics.get("Error Rate", 0) == 0



from veeksha.core.lazy_loader import _LazyLoader
from veeksha.types import SessionGeneratorType
from veeksha.types.base_registry import BaseRegistry


class SessionGeneratorRegistry(BaseRegistry):
    @classmethod
    def get_key_from_str(cls, key_str: str) -> SessionGeneratorType:
        return SessionGeneratorType.from_str(key_str)  # type: ignore


SessionGeneratorRegistry.register(
    SessionGeneratorType.SYNTHETIC,
    _LazyLoader(
        "veeksha.generator.session.synthetic",
        "SyntheticSessionGenerator",
    ),
)
SessionGeneratorRegistry.register(
    SessionGeneratorType.LMEVAL,
    _LazyLoader(
        "veeksha.generator.session.lmeval",
        "LMEvalSessionGenerator",
    ),
)
SessionGeneratorRegistry.register(
    SessionGeneratorType.TRACE,
    _LazyLoader(
        "veeksha.generator.session.trace",
        "TraceSessionGenerator",
    ),
)



import json
import queue
import threading
import time
from typing import Any, Dict, List, Optional

from veeksha.config.trace_recorder import TraceRecorderConfig
from veeksha.logger import init_logger

logger = init_logger(__name__)


class TraceRecorder:
    """Records dispatched requests to a JSONL trace file asynchronously."""

    def __init__(
        self,
        output_dir: str,
        benchmark_start_time: float,
        config: TraceRecorderConfig,
    ):
        """Initialize the trace recorder.

        Args:
            config: Trace recorder configuration.
        """
        self.config = config
        self.output_dir = output_dir
        self.benchmark_start_time = benchmark_start_time
        self.include_content = self.config.include_content
        self.trace_file_path = f"{self.output_dir}/dispatch_trace.jsonl"

        self._queue: queue.Queue = queue.Queue()
        self._stop_event = threading.Event()
        self._writer_thread: Optional[threading.Thread] = None
        self._batch_size = 100
        self._flush_interval = 1.0  # seconds

        try:
            with open(self.trace_file_path, "w") as f:
                pass
            logger.info(f"Initialized trace file at {self.trace_file_path}")
        except Exception as e:
            logger.error(f"Failed to initialize trace file: {e}")

    def start(self) -> None:
        """Start the background writer thread."""
        if self._writer_thread is not None:
            return

        self._stop_event.clear()
        self._writer_thread = threading.Thread(
            target=self._writer_loop, name="TraceRecorderWriter", daemon=True
        )
        self._writer_thread.start()
        logger.info("Trace recorder writer thread started")

    def stop(self) -> None:
        """Stop the writer thread and flush remaining items."""
        if self._writer_thread is None:
            return

        logger.info("Stopping trace recorder, flushing remaining items...")
        self._stop_event.set()
        self._writer_thread.join(timeout=5.0)
        self._writer_thread = None

    def record_dispatch(
        self,
        request: Any,
        session_id: int,
        session_size: int,
        dispatched_at: float,
    ) -> None:
        """Queue a dispatched request for recording.

        Args:
            request: The dispatched request object
            session_id: Session ID
            session_size: Total requests in the session
            dispatched_at: Monotonic timestamp of dispatch
        """
        try:
            channels_data = None
            history_data = None

            trace_entry = {
                "request_id": request.id,
                "session_id": session_id,
                "session_size": session_size,
                "dispatched_at": round(dispatched_at - self.benchmark_start_time, 5),
                "session_context": request.session_context,
            }

            if self.include_content:
                channels_data = {
                    str(modality.name).lower(): self._serialize_channel_content(content)
                    for modality, content in request.channels.items()
                }
                trace_entry["channels"] = channels_data
                trace_entry["history"] = request.history

            self._queue.put(trace_entry)

        except Exception as e:
            logger.error(f"Failed to queue trace for request {request.id}: {e}")

    def _writer_loop(self) -> None:
        """Background loop to write traces to file."""
        buffer: List[Dict[str, Any]] = []
        last_flush_time = time.monotonic()

        with open(self.trace_file_path, "a") as f:
            while not self._stop_event.is_set() or not self._queue.empty():
                try:
                    timeout = 0.1
                    try:
                        item = self._queue.get(timeout=timeout)
                        buffer.append(item)
                    except queue.Empty:
                        pass

                    current_time = time.monotonic()
                    should_flush = (
                        len(buffer) >= self._batch_size
                        or (
                            buffer
                            and (current_time - last_flush_time) >= self._flush_interval
                        )
                        or (self._stop_event.is_set() and buffer)
                    )

                    if should_flush:
                        self._flush_buffer(f, buffer)
                        buffer.clear()
                        last_flush_time = current_time

                except Exception as e:
                    logger.error(f"Error in trace writer loop: {e}")
                    buffer.clear()

    def _flush_buffer(self, f, buffer: List[Dict[str, Any]]) -> None:
        """Write buffered items to file."""
        try:
            lines = [json.dumps(entry) + "\n" for entry in buffer]
            f.writelines(lines)
            f.flush()
        except Exception as e:
            logger.error(f"Failed to write batch of {len(buffer)} traces: {e}")

    def _serialize_channel_content(self, content: Any) -> Dict[str, Any]:
        """Serialize channel content to a dictionary."""
        if hasattr(content, "__dataclass_fields__"):
            from dataclasses import asdict

            return asdict(content)
        try:
            return vars(content)
        except TypeError:
            return {"raw_str": str(content)}



from veeksha.slo.slo import ConstantSlo
from veeksha.types import SloType
from veeksha.types.base_registry import BaseRegistry


class SloRegistry(BaseRegistry):
    """Registry mapping SLO config types to runtime SLO evaluators."""

    @classmethod
    def get_key_from_str(cls, key_str: str) -> SloType:
        return SloType.from_str(key_str)  # type: ignore


SloRegistry.register(SloType.CONSTANT, ConstantSlo)
